#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2020 Satpy developers
#
# satpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# satpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with satpy.  If not, see <http://www.gnu.org/licenses/>.

"""The ici_l1b_nc reader tests package.

This version tests the reader for ICI test data as per PFS V3A.

"""

import datetime as dt
from unittest.mock import patch

import numpy as np
import pytest
import xarray as xr
from netCDF4 import Dataset

from satpy.readers.ici_l1b_nc import IciL1bNCFileHandler, InterpolationType

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - tmp_path

N_CHANNELS = 13
N_SCAN = 10
N_SAMPLES = 784
N_SUBS = 158
N_HORNS = 7
N_183 = 3


@pytest.fixture
def reader(fake_file):
    """Return reader of ici level1b data."""
    return IciL1bNCFileHandler(
        filename=fake_file,
        filename_info={
            "sensing_start_time": (
                dt.datetime.fromisoformat("2000-01-01T01:00:00")
            ),
            "sensing_end_time": (
                dt.datetime.fromisoformat("2000-01-01T02:00:00")
            ),
            "creation_time": (
                dt.datetime.fromisoformat("2000-01-01T03:00:00")
            ),
        },
        filetype_info={
            "longitude": "data/navigation_data/longitude",
            "latitude": "data/navigation_data/latitude",
            "solar_azimuth": "data/navigation_data/ici_solar_azimuth_angle",
            "solar_zenith": "data/navigation_data/ici_solar_zenith_angle",
        }
    )


@pytest.fixture
def fake_file(tmp_path):
    """Return file path to level1b file."""
    file_path = tmp_path / "test_file_ici_l1b_nc.nc"
    writer = IciL1bFakeFileWriter(file_path)
    writer.write()
    return file_path


@pytest.fixture
def dataset_info():
    """Return dataset info."""
    return {
        "name": "1",
        "file_type": "nc_ici_l1b_rad",
        "file_key": "data/measurement_data/ici_radiance_183",
        "coordinates": ["lat_pixels_horn_1", "lon_pixels_horn_1"],
        "n_183": 0,
        "chan_index": 0,
        "calibration": "brightness_temperature",
    }


class IciL1bFakeFileWriter:
    """Writer class of fake ici level1b data."""

    def __init__(self, file_path):
        """Init."""
        self.file_path = file_path

    def write(self):
        """Write fake data to file."""
        with Dataset(self.file_path, "w") as dataset:
            self._write_attributes(dataset)
            self._write_quality_group(dataset)
            data_group = dataset.createGroup("data")
            self._write_measurement_data_group(data_group)
            self._write_navigation_data_group(data_group)

    @staticmethod
    def _write_attributes(dataset):
        """Write attributes."""
        dataset.sensing_start_time_utc = "2000-01-02 03:04:05.000"
        dataset.sensing_end_time_utc = "2000-01-02 04:05:06.000"
        dataset.instrument = "ICI"
        dataset.spacecraft = "SGB"

    @staticmethod
    def _write_quality_group(dataset):
        """Write the quality group."""
        group = dataset.createGroup("quality")
        group.overall_quality_flag = 0
        duration_of_product = group.createVariable(
            "duration_of_product", "f4"
        )
        duration_of_product[:] = 1000.

    @staticmethod
    def _write_navigation_data_group(dataset):
        """Write the navigation data group."""
        group = dataset.createGroup("navigation_data")
        group.createDimension("n_scan", N_SCAN)
        group.createDimension("n_samples", N_SAMPLES)
        group.createDimension("n_subs", N_SUBS)
        group.createDimension("n_horns", N_HORNS)
        subs = group.createVariable("n_subs", "i4", dimensions=("n_subs",))
        subs[:] = np.arange(N_SUBS)
        dimensions = ("n_scan", "n_subs", "n_horns")
        shape = (N_SCAN, N_SUBS, N_HORNS)
        longitude = group.createVariable(
            "longitude",
            np.float32,
            dimensions=dimensions,
        )
        longitude[:] = np.ones(shape)
        latitude = group.createVariable(
            "latitude",
            np.float32,
            dimensions=dimensions,
        )
        latitude[:] = 2. * np.ones(shape)
        azimuth = group.createVariable(
            "ici_solar_azimuth_angle",
            np.float32,
            dimensions=dimensions,
        )
        azimuth[:] = 3. * np.ones(shape)
        zenith = group.createVariable(
            "ici_solar_zenith_angle",
            np.float32,
            dimensions=dimensions,
        )
        zenith[:] = 4. * np.ones(shape)
        dimensions = ("n_scan", "n_samples", "n_horns")
        shape = (N_SCAN, N_SAMPLES, N_HORNS)
        delta_longitude = group.createVariable(
            "delta_longitude",
            np.float32,
            dimensions=dimensions,
        )
        delta_longitude[:] = 1000. * np.ones(shape)
        delta_latitude = group.createVariable(
            "delta_latitude",
            np.float32,
            dimensions=dimensions,
        )
        delta_latitude[:] = 1000. * np.ones(shape)

    @staticmethod
    def _write_measurement_data_group(dataset):
        """Write the measurement data group."""
        group = dataset.createGroup("measurement_data")
        group.createDimension("n_scan", N_SCAN)
        group.createDimension("n_samples", N_SAMPLES)
        group.createDimension("n_channels", N_CHANNELS)
        group.createDimension("n_183", N_183)
        scan = group.createVariable("n_scan", "i4", dimensions=("n_scan",))
        scan[:] = np.arange(N_SCAN)
        samples = group.createVariable(
            "n_samples", "i4", dimensions=("n_samples",)
        )
        samples[:] = np.arange(N_SAMPLES)
        bt_a = group.createVariable(
            "bt_conversion_a", np.float32, dimensions=("n_channels",)
        )
        bt_a[:] = np.ones(N_CHANNELS)
        bt_b = group.createVariable(
            "bt_conversion_b", np.float32, dimensions=("n_channels",)
        )
        bt_b[:] = np.zeros(N_CHANNELS)
        cw = group.createVariable(
            "centre_wavenumber", np.float32, dimensions=("n_channels",)
        )
        cw[:] = np.array(
            [6.0] * 3 + [8.0] * 2 + [11.0] * 3 + [15.0] * 3 + [22.0] * 2
        )
        ici_radiance_183 = group.createVariable(
            "ici_radiance_183",
            np.float32,
            dimensions=("n_scan", "n_samples", "n_183"),
        )
        ici_radiance_183[:] = 0.08 * np.ones((N_SCAN, N_SAMPLES, N_183))


class TestIciL1bNCFileHandler:
    """Test the IciL1bNCFileHandler reader."""

    def test_start_time(self, reader):
        """Test start time."""
        assert reader.start_time == dt.datetime(2000, 1, 2, 3, 4, 5)

    def test_end_time(self, reader):
        """Test end time."""
        assert reader.end_time == dt.datetime(2000, 1, 2, 4, 5, 6)

    def test_sensor(self, reader):
        """Test sensor."""
        assert reader.sensor == "ICI"

    def test_platform_name(self, reader):
        """Test platform name."""
        assert reader.platform_name == "SGB"

    def test_ssp_lon(self, reader):
        """Test sub satellite path longitude."""
        assert reader.ssp_lon is None

    def test_longitude(self, reader):
        """Test longitude."""
        np.testing.assert_allclose(reader.longitude, 1, rtol=1e-3)

    def test_latitude(self, reader):
        """Test latitude."""
        np.testing.assert_allclose(reader.latitude, 2, rtol=1e-3)

    def test_solar_azimuth(self, reader):
        """Test solar azimuth."""
        np.testing.assert_allclose(reader.solar_azimuth, 3, rtol=1e-3)

    def test_solar_zenith(self, reader):
        """Test solar zenith."""
        np.testing.assert_allclose(reader.solar_zenith, 4, rtol=1e-3)

    def test_calibrate_raises_for_unknown_calibration_method(self, reader):
        """Test perform calibration raises for unknown calibration method."""
        variable = xr.DataArray(np.ones(3))
        dataset_info = {"calibration": "unknown", "name": "radiance"}
        with pytest.raises(ValueError, match="Unknown calibration"):
            reader._calibrate(variable, dataset_info)

    @patch("satpy.readers.ici_l1b_nc.IciL1bNCFileHandler._calibrate_bt")
    def test_calibrate_does_not_call_calibrate_bt_if_not_needed(
        self,
        mocked_calibrate,
        reader,
    ):
        """Test calibrate does not call calibrate_bt if not needed."""
        variable = xr.DataArray(
            np.array([
                [0.060, 0.065, 0.070, 0.075],
                [0.080, 0.085, 0.090, 0.095],
            ]),
            dims=("n_scan", "n_samples"),
        )
        dataset_info = {"calibration": "radiance"}
        reader._calibrate(variable, dataset_info)
        mocked_calibrate.assert_not_called()

    @patch("satpy.readers.ici_l1b_nc.IciL1bNCFileHandler._calibrate_bt")
    def test_calibrate_calls_calibrate_bt(
        self,
        mocked_calibrate_bt,
        reader,
    ):
        """Test calibrate calls calibrate_bt."""
        variable = xr.DataArray(
            np.array([
                [0.060, 0.065, 0.070, 0.075],
                [0.080, 0.085, 0.090, 0.095],
            ]),
            dims=("n_scan", "n_samples"),
        )
        dataset_info = {
            "calibration": "brightness_temperature",
            "chan_index": 2,
        }
        reader._calibrate(variable, dataset_info)
        mocked_calibrate_bt.assert_called_once_with(
            variable,
            6.0,
            1.0,
            0.0,
        )

    def test_calibrate_bt(self, reader):
        """Test calibrate brightness temperature."""
        radiance = xr.DataArray(
            np.array([
                [0.060, 0.065, 0.070, 0.075],
                [0.080, 0.085, 0.090, 0.095],
            ])
        )
        cw = 6.1145
        a = 1.
        b = 0.0
        bt = reader._calibrate_bt(radiance, cw, a, b)
        expected_bt = np.array([
            [198.22929022, 214.38700287, 230.54437184, 246.70146465],
            [262.85833223, 279.01501371, 295.17153966, 311.32793429],
        ])
        np.testing.assert_allclose(bt, expected_bt)

    @pytest.mark.parametrize("dims", [
        ("n_scan", "n_samples"),
        ("x", "y"),
    ])
    def test_standardize_dims(self, reader, dims):
        """Test standardize dims."""
        variable = xr.DataArray(
            np.arange(6).reshape(2, 3),
            dims=dims,
        )
        standardized = reader._standardize_dims(variable)
        assert standardized.dims == ("y", "x")

    @pytest.mark.parametrize(("dims", "data_info", "expect"), [
        (("y", "x", "n_horns"), {"n_horns": 1}, 1),
        (("y", "x", "n_183"), {"n_183": 2}, 2),
    ])
    def test_filter_variable(self, reader, dims, data_info, expect):
        """Test filter variable."""
        data = np.arange(24).reshape(2, 3, 4)
        variable = xr.DataArray(
            np.arange(24).reshape(2, 3, 4),
            dims=dims,
        )
        filtered = reader._filter_variable(variable, data_info)
        assert filtered.dims == ("y", "x")
        assert (filtered == data[:, :, expect]).all()

    def test_drop_coords(self, reader):
        """Test drop coordinates."""
        coords = "dummy"
        data = xr.DataArray(
            np.ones(10),
            dims=("y"),
            coords={coords: 0},
        )
        assert coords in data.coords
        data = reader._drop_coords(data)
        assert coords not in data.coords

    def test_get_third_dimension_name(self, reader):
        """Test get third dimension name."""
        data = xr.DataArray(np.ones((1, 1, 1)), dims=("x", "y", "z"))
        assert reader._get_third_dimension_name(data) == "z"

    def test_get_third_dimension_name_return_none_for_2d_data(self, reader):
        """Test get third dimension name return none for 2d data."""
        data = xr.DataArray(np.ones((1, 1)), dims=("x", "y"))
        assert reader._get_third_dimension_name(data) is None

    def test_get_dataset_return_none_if_data_not_exist(self, reader):
        """Tes get dataset return none if data does not exist."""
        dataset_id = {"name": "unknown"}
        dataset_info = {"file_key": "non/existing/data"}
        dataset = reader.get_dataset(dataset_id, dataset_info)
        assert dataset is None

    @patch("satpy.readers.ici_l1b_nc.IciL1bNCFileHandler._calibrate_bt")
    def test_get_dataset_does_not_calibrate_if_not_desired(
        self,
        mocked_calibrate,
        reader,
        dataset_info,
    ):
        """Test get dataset does not calibrate if not desired."""
        dataset_id = {"name": "1"}
        dataset_info.pop("calibration")
        dataset = reader.get_dataset(dataset_id, dataset_info)
        assert dataset.dims == ("y", "x")
        mocked_calibrate.assert_not_called()
        assert isinstance(dataset, xr.DataArray)

    def test_get_dataset_orthorectifies_if_orthorect_data_defined(
        self,
        reader,
    ):
        """Test get dataset orthorectifies if orthorect data is defined."""
        dataset_id = {"name": "lon_pixels_horn_1"}
        dataset_info = {
            "name": "lon_pixels_horn_1",
            "file_type": "nc_ici_l1b_rad",
            "file_key": "longitude",
            "orthorect_data": "data/navigation_data/delta_longitude",
            "standard_name": "longitude",
            "n_horns": 0,
            "modifiers": (),
        }
        dataset = reader.get_dataset(dataset_id, dataset_info)
        np.testing.assert_allclose(dataset, 1.009139, atol=1e-6)

    def test_get_dataset_handles_calibration(
        self,
        reader,
        dataset_info,
    ):
        """Test get dataset handles calibration."""
        dataset_id = {"name": "1"}
        dataset = reader.get_dataset(dataset_id, dataset_info)
        assert dataset.attrs["calibration"] == "brightness_temperature"
        np.testing.assert_allclose(dataset, 272.73734)

    def test_interpolate_returns_none_if_dataset_not_exist(self, reader):
        """Test interpolate returns none if dataset not exist."""
        azimuth, zenith = reader._interpolate(
            InterpolationType.OBSERVATION_ANGLES
        )
        assert azimuth is None
        assert zenith is None

    @patch("satpy.readers.ici_l1b_nc.IciL1bNCFileHandler._interpolate_geo")
    def test_interpolate_calls_interpolate_geo(self, mock, reader):
        """Test interpolate calls interpolate_geo."""
        reader._interpolate(InterpolationType.LONLAT)
        mock.assert_called_once()

    @patch("satpy.readers.ici_l1b_nc.IciL1bNCFileHandler._interpolate_viewing_angle")  # noqa: E501
    def test_interpolate_calls_interpolate_viewing_angles(self, mock, reader):
        """Test interpolate calls interpolate viewing_angles."""
        reader._interpolate(InterpolationType.SOLAR_ANGLES)
        mock.assert_called_once()

    def test_interpolate_geo(self, reader):
        """Test interpolate geographic coordinates."""
        shape = (N_SCAN, N_SUBS, N_HORNS)
        dims = ("n_scan", "n_subs", "n_horns")
        sub_pos = np.append(
            np.arange(0, N_SAMPLES, np.ceil(N_SAMPLES / N_SUBS)),
            N_SAMPLES - 1
        )
        longitude = xr.DataArray(
            np.tile( # longitudes between 0 and 10
                10 * sub_pos / sub_pos[-1], (N_SCAN, N_HORNS, 1)
            ).swapaxes(1, 2),
            dims=dims,
            coords={
                "n_horns": np.arange(N_HORNS),
                "n_subs": np.arange(N_SUBS),
            },
        )
        latitude = xr.DataArray(np.ones(shape), dims=dims)
        lon, lat = reader._interpolate_geo(
            longitude,
            latitude,
            N_SAMPLES,
        )
        expect_shape = (N_SCAN, N_SAMPLES, N_HORNS)
        assert lon.shape == expect_shape
        assert lat.shape == expect_shape
        np.testing.assert_allclose(lon[:, 0, :], 0.)
        np.testing.assert_allclose(lon[:, -1, :], 10.)
        np.testing.assert_allclose(np.diff(lon[0, :, 0]), 10 / (N_SAMPLES - 1))
        np.testing.assert_allclose(lat, 1.0)

    def test_interpolate_viewing_angle(self, reader):
        """Test interpolate viewing angle."""
        shape = (N_SCAN, N_SUBS, N_HORNS)
        dims = ("n_scan", "n_subs", "n_horns")
        azimuth = xr.DataArray(
            np.ones(shape),
            dims=dims,
            coords={
                "n_horns": np.arange(N_HORNS),
                "n_subs": np.arange(N_SUBS),
            },
        )
        zenith = xr.DataArray(100. * np.ones(shape), dims=dims)
        azimuth, zenith = reader._interpolate_viewing_angle(
            azimuth,
            zenith,
            N_SAMPLES,
        )
        expect_shape = (N_SCAN, N_SAMPLES, N_HORNS)
        assert azimuth.shape == expect_shape
        assert zenith.shape == expect_shape
        np.testing.assert_allclose(azimuth, 1.0)
        np.testing.assert_allclose(zenith, 100.0)

    def test_orthorectify(self, reader):
        """Test orthorectify."""
        variable = xr.DataArray(
            np.ones((N_SCAN, N_SAMPLES, N_HORNS)),
            dims=("y", "x", "n_horns"),
            coords={"n_horns": np.arange(N_HORNS)}
        )
        variable = variable.sel({"n_horns": 0})
        orthorect_data_name = "data/navigation_data/delta_longitude"
        orthorectified = reader._orthorectify(
            variable,
            orthorect_data_name,
        )
        np.testing.assert_allclose(orthorectified, 1.009, rtol=1e-5)

    def test_get_global_attributes(self, reader):
        """Test get global attributes."""
        attributes = reader._get_global_attributes()
        assert attributes == {
            "filename": reader.filename,
            "start_time": dt.datetime(2000, 1, 2, 3, 4, 5),
            "end_time": dt.datetime(2000, 1, 2, 4, 5, 6),
            "spacecraft_name": "SGB",
            "ssp_lon": None,
            "sensor": "ICI",
            "filename_start_time": dt.datetime(2000, 1, 1, 1, 0),
            "filename_end_time": dt.datetime(2000, 1, 1, 2, 0),
            "platform_name": "SGB",
            "quality_group": {
                "duration_of_product": np.array(1000., dtype=np.float32),
                "overall_quality_flag": 0,
            }
        }

    def test_get_quality_attributes(self, reader):
        """Test get quality attributes."""
        attributes = reader._get_quality_attributes()
        assert attributes == {
            "duration_of_product": np.array(1000., dtype=np.float32),
            "overall_quality_flag": 0,
        }

    @patch(
        "satpy.readers.ici_l1b_nc.IciL1bNCFileHandler._get_global_attributes",
        return_value={"mocked_global_attributes": True},
    )
    def test_manage_attributes(self, mock, reader):
        """Test manage attributes."""
        variable = xr.DataArray(
            np.ones(N_SCAN),
            attrs={"season": "summer"},
        )
        dataset_info = {"name": "ici_1", "units": "K"}
        variable = reader._manage_attributes(variable, dataset_info)
        assert variable.attrs == {
            "season": "summer",
            "units": "K",
            "name": "ici_1",
            "mocked_global_attributes": True,
        }



# Copyright (c) 2015-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Composites."""

from __future__ import annotations

from typing import Any

from satpy.utils import _import_and_warn_new_location

IMPORT_PATHS = {
    "DifferenceCompositor": "satpy.composites.arithmetic",
    "RatioCompositor": "satpy.composites.arithmetic",
    "SumCompositor": "satpy.composites.arithmetic",
    "StaticImageCompositor": "satpy.composites.aux_data",
    "IncompatibleAreas": "satpy.composites.core",
    "IncompatibleTimes": "satpy.composites.core",
    "CompositeBase": "satpy.composites.core",
    "GenericCompositor": "satpy.composites.core",
    "RGBCompositor": "satpy.composites.core",
    "SingleBandCompositor": "satpy.composites.core",
    "add_bands": "satpy.composites.core",
    "check_times": "satpy.composites.core",
    "enhance2dataset": "satpy.composites.core",
    "BackgroundCompositor": "satpy.composites.fill",
    "DayNightCompositor": "satpy.composites.fill",
    "Filler": "satpy.composites.fill",
    "FillingCompositor": "satpy.composites.fill",
    "MultiFiller": "satpy.composites.fill",
    "add_alpha_bands": "satpy.composites.fill",
    "zero_missing_data": "satpy.composites.fill",
    "CategoricalDataCompositor": "satpy.composites.lookup",
    "ColorizeCompositor": "satpy.composites.lookup",
    "ColormapCompositor": "satpy.composites.lookup",
    "PaletteCompositor": "satpy.composites.lookup",
    "CloudCompositor": "satpy.composites.mask",
    "HighCloudCompositor": "satpy.composites.mask",
    "LongitudeMaskingCompositor": "satpy.composites.mask",
    "LowCloudCompositor": "satpy.composites.mask",
    "MaskingCompositor": "satpy.composites.mask",
    "SimpleFireMaskCompositor": "satpy.composites.mask",
    "RealisticColors": "satpy.composites.seviri",
    "LuminanceSharpeningCompositor": "satpy.composites.resolution",
    "RatioSharpenedRGB": "satpy.composites.resolution",
    "SandwichCompositor": "satpy.composites.resolution",
    "SelfSharpenedRGB": "satpy.composites.resolution",
    "NaturalEnh": "satpy.composites.spectral",
}


def __getattr__(name: str) -> Any:
    new_module = IMPORT_PATHS.get(name)

    if new_module is None:
        raise AttributeError(f"module {__name__} has no attribute '{name}'")

    return _import_and_warn_new_location(new_module, name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2020 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Reader for AMSR2 L2 files in HDF5 format."""

from satpy.readers.amsr2_l1b import AMSR2L1BFileHandler


class AMSR2L2FileHandler(AMSR2L1BFileHandler):
    """AMSR2 level 2 file handler."""

    def mask_dataset(self, ds_info, data):
        """Mask data with the fill value."""
        fill_value = ds_info.get("fill_value", 65535)
        return data.where(data != fill_value)

    def scale_dataset(self, var_path, data):
        """Scale data with the scale factor attribute."""
        return data * self[var_path + "/attr/SCALE FACTOR"]

    def get_dataset(self, ds_id, ds_info):
        """Get output data and metadata of specified dataset."""
        var_path = ds_info["file_key"]

        data = self[var_path].squeeze()
        data = self.mask_dataset(ds_info, data)
        data = self.scale_dataset(var_path, data)

        if ds_info.get("name") == "ssw":
            data = data.rename({"dim_0": "y", "dim_1": "x"})
        metadata = self.get_metadata(ds_id, ds_info)
        data.attrs.update(metadata)
        return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Geostationary Projection / Area computations.

This module computes properties and area definitions for geostationary
satellites. It is designed to be a common module that can be called by
all geostationary satellite readers and uses commonly-included parameters
such as the CFAC/LFAC values, satellite position, etc, to compute the
correct area definition.
"""

import numpy as np
from pyresample import geometry


def get_xy_from_linecol(line, col, offsets, factors):
    """Get the intermediate coordinates from line & col.

    Intermediate coordinates are actually the instruments scanning angles.
    """
    loff, coff = offsets
    lfac, cfac = factors
    x__ = float(col - coff) / (float(cfac) / 2 ** 16)
    y__ = float(line - loff) / (float(lfac) / 2 ** 16)

    return x__, y__


def make_ext(ll_x, ur_x, ll_y, ur_y, h):
    """Create the area extent from computed ll and ur.

    Args:
        ll_x (numbers.Number): The lower left x coordinate (m)
        ur_x (numbers.Number): The upper right x coordinate (m)
        ll_y (numbers.Number): The lower left y coordinate (m)
        ur_y (numbers.Number): The upper right y coordinate (m)
        h (numbers.Number): The satellite altitude above the Earth's surface
    Returns:
        (tuple[numbers.Number]): An area extent for the scene

    """
    aex = (np.deg2rad(ll_x) * h, np.deg2rad(ll_y) * h,
           np.deg2rad(ur_x) * h, np.deg2rad(ur_y) * h)

    return aex


def get_area_extent(pdict):
    """Get the area extent seen by a geostationary satellite.

    Args:
        pdict (dict): A dictionary containing common parameters:
            nlines (numbers.Number): Number of lines in image
            ncols (numbers.Number): Number of columns in image
            cfac (numbers.Number): Column scaling factor
            lfac (numbers.Number): Line scaling factor
            coff (numbers.Number): Column offset factor
            loff (numbers.Number): Line offset factor
            scandir(str): 'N2S' for standard (N->S), 'S2N' for inverse (S->N)
            h (numbers.Number): Altitude of satellite above the Earth's surface (m)

    Returns:
        (tuple[numbers.Number]): An area extent for the scene

    """
    # count starts at 1
    cols = 1 - 0.5

    if pdict["scandir"] == "S2N":
        lines = 0.5 - 1
        scanmult = -1
    else:
        lines = 1 - 0.5
        scanmult = 1
    # Lower left x, y scanning angles in degrees
    ll_x, ll_y = get_xy_from_linecol(lines * scanmult,
                                     cols,
                                     (pdict["loff"], pdict["coff"]),
                                     (pdict["lfac"], pdict["cfac"]))

    cols += pdict["ncols"]
    lines += pdict["nlines"]
    # Upper right x, y scanning angles in degrees
    ur_x, ur_y = get_xy_from_linecol(lines * scanmult,
                                     cols,
                                     (pdict["loff"], pdict["coff"]),
                                     (pdict["lfac"], pdict["cfac"]))
    if pdict["scandir"] == "S2N":
        ll_y *= -1
        ur_y *= -1

    # Convert degrees to radians and create area extent
    aex = make_ext(ll_x=ll_x, ur_x=ur_x, ll_y=ll_y, ur_y=ur_y, h=pdict["h"])

    return aex


def get_area_definition(pdict, a_ext):
    """Get the area definition for a geo-sat.

    Args:
        pdict: A dictionary containing common parameters:
            nlines: Number of lines in image
            ncols: Number of columns in image
            ssp_lon: Subsatellite point longitude (deg)
            a: Earth equatorial radius (m)
            b: Earth polar radius (m)
            h: Platform height (m)
            a_name: Area name
            a_desc: Area description
            p_id: Projection id
        a_ext: A four element tuple containing the area extent (scan angle)
               for the scene in radians
    Returns:
        An area definition for the scene

    .. note::

        The AreaDefinition `proj_id` attribute is being deprecated.

    """
    proj_dict = {"a": float(pdict["a"]),
                 "b": float(pdict["b"]),
                 "lon_0": float(pdict["ssp_lon"]),
                 "h": float(pdict["h"]),
                 "proj": "geos",
                 "units": "m"}

    a_def = geometry.AreaDefinition(
        pdict["a_name"],
        pdict["a_desc"],
        pdict["p_id"],
        proj_dict,
        int(pdict["ncols"]),
        int(pdict["nlines"]),
        a_ext)

    return a_def


def sampling_to_lfac_cfac(sampling):
    """Convert angular sampling to line/column scaling factor (aka LFAC/CFAC).

    Reference: `MSG Ground Segment LRIT HRIT Mission Specific Implementation`_,
    Appendix E.2.

    .. _MSG Ground Segment LRIT HRIT Mission Specific Implementation:
       https://user.eumetsat.int/s3/eup-strapi-media/pdf_ten_05057_spe_msg_lrit_hri_bcaff5f0ba.pdf

    Args:
        sampling: float
            Angular sampling (rad)

    Returns:
        Line/column scaling factor (deg-1)

    """
    return 2.0 ** 16 / np.rad2deg(sampling)


def get_geos_area_naming(input_dict):
    """Get a dictionary containing formatted AreaDefinition naming.

    Args:
        input_dict: dict
            Dictionary with keys `platform_name`, `instrument_name`, `service_name`, `service_desc`, `resolution` .
            The resolution is expected in meters.

    Returns:
        area_naming_dict with `area_id`, `description`  keys, values are strings.

    .. note::

        The AreaDefinition `proj_id` attribute is being deprecated and is therefore not formatted here.
        An empty string is to be used until the attribute is fully removed.

    """
    area_naming_dict = {}

    resolution_strings = get_resolution_and_unit_strings(input_dict["resolution"])

    area_naming_dict["area_id"] = "{}_{}_{}_{}{}".format(input_dict["platform_name"].lower(),
                                                         input_dict["instrument_name"].lower(),
                                                         input_dict["service_name"].lower(),
                                                         resolution_strings["value"],
                                                         resolution_strings["unit"]
                                                         )

    area_naming_dict["description"] = "{} {} {} area definition " \
                                      "with {} {} resolution".format(input_dict["platform_name"].upper(),
                                                                     input_dict["instrument_name"].upper(),
                                                                     input_dict["service_desc"],
                                                                     resolution_strings["value"],
                                                                     resolution_strings["unit"]
                                                                     )

    return area_naming_dict


def get_resolution_and_unit_strings(resolution):
    """Get the resolution value and unit as strings.

    If the resolution is larger than 1000 m, use kilometer as unit. If lower, use meter.

    Args:
        resolution: scalar
            Resolution in meters.

    Returns:
        Dictionary with `value` and `unit` keys, values are strings.
    """
    if resolution >= 1000:
        return {"value": "{:.0f}".format(resolution*1e-3),
                "unit": "km"}

    return {"value": "{:.0f}".format(resolution),
            "unit": "m"}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Module for testing the satpy.readers.grib module."""

import datetime as dt
import sys
import unittest
from unittest import mock

import numpy as np
from pytest import approx  # noqa: PT013

from satpy.tests.utils import make_dataid


class FakeMessage(object):
    """Fake message returned by pygrib.open().message(x)."""

    def __init__(self, values, proj_params=None, latlons=None, **attrs):
        """Init the fake message."""
        super(FakeMessage, self).__init__()
        self.attrs = attrs
        self.values = values
        if proj_params is None:
            proj_params = {"a": 6378140.0, "b": 6356755.0, "lat_0": 0.0,
                           "lon_0": 0.0, "proj": "geos", "h": 35785830.098}
        self.projparams = proj_params
        self._latlons = latlons

    def latlons(self):
        """Get the latlons."""
        return self._latlons

    def __getitem__(self, item):
        """Get item."""
        return self.attrs[item]

    def valid_key(self, key):
        """Check if key is valid."""
        return True


class FakeGRIB(object):
    """Fake GRIB file returned by pygrib.open."""

    def __init__(self, messages=None, proj_params=None, latlons=None):
        """Init the fake grib file."""
        super(FakeGRIB, self).__init__()
        if messages is not None:
            self._messages = messages
        else:
            self._messages = [
                FakeMessage(
                    values=np.arange(25.).reshape((5, 5)),
                    name="Instantaneous rain rate",
                    shortName="irrate",
                    cfName="unknown",
                    units="kg m**-2 s**-1",
                    dataDate=20190603,
                    dataTime=1645,
                    missingValue=9999,
                    modelName="unknown",
                    centreDescription="Rome",
                    minimum=0.0,
                    maximum=0.01475,
                    Nx=3712,
                    Ny=3712,
                    NrInRadiusOfEarth=6.6107,
                    dx=3622,
                    dy=3610,
                    XpInGridLengths=1856.0,
                    YpInGridLengths=1856.0,
                    jScansPositively=0,
                    proj_params=proj_params,
                    latlons=latlons,
                )
            ]
        self.messages = len(self._messages)

    def message(self, msg_num):
        """Fake message."""
        return self._messages[msg_num - 1]

    def seek(self, loc):
        """Fake seek."""
        return

    def __iter__(self):
        """Iterate over messages."""
        return iter(self._messages)

    def __enter__(self):
        """Enter the context."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit the context."""


class TestHSAFFileHandler(unittest.TestCase):
    """Test HSAF Reader."""

    def setUp(self):
        """Wrap pygrib to read fake data."""
        try:
            import pygrib
        except ImportError:
            pygrib = None
        self.orig_pygrib = pygrib
        sys.modules["pygrib"] = mock.MagicMock()

    def tearDown(self):
        """Re-enable pygrib import."""
        sys.modules["pygrib"] = self.orig_pygrib

    @mock.patch("satpy.readers.hsaf_grib.pygrib.open", return_value=FakeGRIB())
    def test_init(self, pg):
        """Test the init function, ensure that the correct dates and metadata are returned."""
        pg.open.return_value = FakeGRIB()
        correct_dt = dt.datetime(2019, 6, 3, 16, 45, 0)
        from satpy.readers.hsaf_grib import HSAFFileHandler
        fh = HSAFFileHandler("filename", mock.MagicMock(), mock.MagicMock())
        assert fh._analysis_time == correct_dt
        assert fh.metadata["projparams"]["lat_0"] == 0.0
        assert fh.metadata["shortName"] == "irrate"
        assert fh.metadata["nx"] == 3712

    @mock.patch("satpy.readers.hsaf_grib.pygrib.open", return_value=FakeGRIB())
    def test_get_area_def(self, pg):
        """Test the area definition setup, checks the size and extent."""
        pg.open.return_value = FakeGRIB()
        from satpy.readers.hsaf_grib import HSAFFileHandler
        fh = HSAFFileHandler("filename", mock.MagicMock(), mock.MagicMock())
        area_def = HSAFFileHandler.get_area_def(fh, "H03B")
        assert area_def.width == 3712
        assert area_def.area_extent[0] == approx(-5569209.3026, abs=1e-3)
        assert area_def.area_extent[3] == approx(5587721.9097, abs=1e-3)

    @mock.patch("satpy.readers.hsaf_grib.pygrib.open", return_value=FakeGRIB())
    def test_get_dataset(self, pg):
        """Test reading the actual datasets from a grib file."""
        pg.open.return_value = FakeGRIB()
        from satpy.readers.hsaf_grib import HSAFFileHandler

        # Instantaneous precipitation
        fh = HSAFFileHandler("filename", mock.MagicMock(), mock.MagicMock())
        fh.filename = "H03B"
        ds_id = make_dataid(name="H03B")
        data = fh.get_dataset(ds_id, mock.Mock())
        np.testing.assert_array_equal(data.values, np.arange(25.).reshape((5, 5)))

        # Accumulated precipitation
        fh = HSAFFileHandler("filename", mock.MagicMock(), mock.MagicMock())
        fh.filename = "H05B"
        ds_id = make_dataid(name="H05B")
        data = fh.get_dataset(ds_id, mock.Mock())
        np.testing.assert_array_equal(data.values, np.arange(25.).reshape((5, 5)))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2021, 2024, 2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Setup and configuration for all reader tests."""

import datetime as dt
import os
from random import randrange

import numpy as np
import pytest
import xarray as xr
from trollsift import compose, parse
from xarray import DataTree

from satpy.readers.mwr_l1b import AWS_EPS_Sterna_MWR_L1BFile
from satpy.readers.mwr_l1c import AWS_MWR_L1CFile

DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S.%f"

platform_name = "AWS1"
# W_XX-EUMETSAT-Darmstadt,SAT,AWS1-MWR-1B-RAD_C_EUMT_20241121085911_G_D_20241109234502_20241110004559_T_N____.nc
file_pattern = "W_{country:2s}-{organisation:s}-{location:s},SAT,{platform_name}-MWR-{processing_level}-RAD_C_{originator:4s}_{processing_time:%Y%m%d%H%M%S}_G_D_{start_time:%Y%m%d%H%M%S}_{end_time:%Y%m%d%H%M%S}_T_B____.nc"  # noqa


rng = np.random.default_rng()

def random_date(start, end):
    """Create a random datetime between two datetimes."""
    delta = end - start
    int_delta = (delta.days * 24 * 60 * 60) + delta.seconds
    random_second = randrange(int_delta)
    return start + dt.timedelta(seconds=random_second)


@pytest.fixture(scope="module")
def fake_mwr_data_array():
    """Return a fake AWS/EPS-Sterna MWR l1b data array."""
    fake_data_np = rng.integers(0, 700000, size=10*145*19).reshape((10, 145, 19))
    fake_data_np[0, 0, 0] = -2147483648
    fake_data_np[1, 0, 0] = 700000 + 10
    fake_data_np[2, 0, 0] = -10
    array_dims = ["n_scans", "n_fovs", "n_channels"]
    return xr.DataArray(fake_data_np, dims=array_dims)


def make_fake_angles(geo_size, geo_dims, shape):
    """Return fake sun-satellite angle array."""
    maxval = 36000
    dummy_array = (np.arange(0, geo_size) * maxval/geo_size).astype("int32")
    return xr.DataArray(dummy_array.reshape(shape), dims=geo_dims)


def make_fake_lsm(geo_size, geo_dims, shape):
    """Return fake land surface type data array."""
    maxval = 1
    dummy_array = (np.arange(0, geo_size) * maxval/geo_size).astype("int16")
    return xr.DataArray(dummy_array.reshape(shape), dims=geo_dims)

def make_fake_dem(geo_size, geo_dims, shape):
    """Return fake DEM data array."""
    maxval = 10000
    dummy_array = (np.arange(0, geo_size) * maxval/geo_size).astype("int16")
    return xr.DataArray(dummy_array.reshape(shape), dims=geo_dims)


def make_fake_mwr_lonlats(geo_size, geo_dims, shape):
    """Return fake geolocation data arrays for all 4 MWR horns."""
    maxval = 3600000
    dummy_array = (np.arange(0, geo_size) * maxval/geo_size).astype("int32")
    fake_lon_data = xr.DataArray(dummy_array.reshape(shape), dims=geo_dims)
    maxval = 1800000
    dummy_array = (np.arange(0, geo_size) * maxval/geo_size - maxval/2).astype("int32")
    fake_lat_data = xr.DataArray(dummy_array.reshape(shape), dims=geo_dims)
    return (fake_lon_data, fake_lat_data)


def make_fake_mwr_l1c_lonlats(geo_size, geo_dims):
    """Return fake level-1c geolocation data arrays."""
    maxval = 3600000
    dummy_array = (np.arange(0, geo_size) * maxval/geo_size).astype("int32")
    fake_lon_data = xr.DataArray(dummy_array.reshape((10, 145)), dims=geo_dims)
    maxval = 1800000
    dummy_array = (np.arange(0, geo_size) * maxval/geo_size - maxval/2).astype("int32")
    fake_lat_data = xr.DataArray(dummy_array.reshape((10, 145)), dims=geo_dims)
    return (fake_lon_data, fake_lat_data)


def aws_eps_sterna_mwr_level1_file(fake_mwr_data_array, eps_sterna=True, l1b=True):
    """Create an AWS and EPS-Sterna MWR l1b file."""
    if eps_sterna:
        n_feedhorns="n_feedhorns"
        prefix = ""
        longitude_attr = "longitude"
        latitude_attr = "latitude"
    else:
        n_feedhorns="n_geo_groups"
        prefix = "aws_"
        longitude_attr = "aws_lon"
        latitude_attr = "aws_lat"

    if l1b:
        geo_dims = ["n_scans", "n_fovs", n_feedhorns]
        geo_size = 10 * 145 * 4
        shape = (10, 145, 4)
    else:
        geo_dims = ["n_scans", "n_fovs"]
        geo_size = 10 * 145
        shape = (10, 145)

    ds = DataTree()
    start_time = dt.datetime(2024, 9, 1, 12, 0)
    ds.attrs["sensing_start_time_utc"] = start_time.strftime(DATETIME_FORMAT)
    end_time = dt.datetime(2024, 9, 1, 12, 15)
    ds.attrs["sensing_end_time_utc"] = end_time.strftime(DATETIME_FORMAT)

    ds.attrs["instrument"] = "MWR"
    ds.attrs["orbit_start"] = 9991
    ds.attrs["orbit_end"] = 9992
    dset_name = f"data/calibration/{prefix}toa_brightness_temperature"
    ds[dset_name] = fake_mwr_data_array
    ds[dset_name].attrs["scale_factor"] = 0.001
    ds[dset_name].attrs["add_offset"] = 0.0
    ds[dset_name].attrs["missing_value"] = -2147483648
    ds[dset_name].attrs["valid_min"] = 0
    ds[dset_name].attrs["valid_max"] = 700000

    fake_lon_data, fake_lat_data = make_fake_mwr_lonlats(geo_size, geo_dims, shape)

    ds[f"data/navigation/{longitude_attr}"] = fake_lon_data
    ds[f"data/navigation/{longitude_attr}"].attrs["scale_factor"] = 1e-4
    ds[f"data/navigation/{longitude_attr}"].attrs["add_offset"] = 0.0
    ds[f"data/navigation/{latitude_attr}"] = fake_lat_data
    ds[f"data/navigation/{prefix}solar_azimuth_angle"] = make_fake_angles(geo_size, geo_dims, shape)
    ds[f"data/navigation/{prefix}solar_zenith_angle"] = make_fake_angles(geo_size, geo_dims, shape)
    ds[f"data/navigation/{prefix}satellite_azimuth_angle"] = make_fake_angles(geo_size, geo_dims, shape)
    ds[f"data/navigation/{prefix}satellite_zenith_angle"] = make_fake_angles(geo_size, geo_dims, shape)
    ds[f"data/navigation/{prefix}surface_type"] = make_fake_lsm(geo_size, geo_dims, shape)
    ds[f"data/navigation/{prefix}terrain_elevation"] = make_fake_dem(geo_size, geo_dims, shape)
    ds[f"data/navigation/{prefix}not_implemented_yet"] = make_fake_dem(geo_size, geo_dims, shape)
    if l1b:
        ds["status/satellite/subsat_latitude_end"] = np.array(22.39)
        ds["status/satellite/subsat_longitude_start"] = np.array(304.79)
        ds["status/satellite/subsat_latitude_start"] = np.array(55.41)
        ds["status/satellite/subsat_longitude_end"] = np.array(296.79)

    return ds


def create_mwr_file(tmpdir, data_array, eps_sterna=False, l1b=True):
    """Create an AWS or EPS-Sterna MWR l1b (or level-1c) file."""
    ds = aws_eps_sterna_mwr_level1_file(data_array, eps_sterna=eps_sterna, l1b=l1b)
    start_time = dt.datetime.fromisoformat(ds.attrs["sensing_start_time_utc"])
    end_time = dt.datetime.fromisoformat(ds.attrs["sensing_end_time_utc"])

    platform_name = "ST01" if eps_sterna else "AWS1"
    processing_level = "1B" if l1b else "1C"

    processing_time = random_date(dt.datetime(2024, 9, 1, 13), dt.datetime(2030, 6, 1))
    filename = tmpdir / compose(file_pattern, dict(country="XX",
                                                   organisation="EUMETSAT",
                                                   location="Darmstadt",
                                                   processing_level=processing_level,
                                                   originator="EUMT",
                                                   start_time=start_time, end_time=end_time,
                                                   processing_time=processing_time,
                                                   platform_name=platform_name))
    ds.to_netcdf(filename)
    return filename

@pytest.fixture(scope="module")
def eps_sterna_mwr_file(tmp_path_factory, fake_mwr_data_array):
    """Create an EPS-Sterna MWR l1b file."""
    tmpdir = tmp_path_factory.mktemp("eps_sterna_mwr_l1b_tests")
    return create_mwr_file(tmpdir, fake_mwr_data_array, eps_sterna=True)


@pytest.fixture(scope="module")
def aws_mwr_file(tmp_path_factory, fake_mwr_data_array):
    """Create an AWS MWR l1b file."""
    tmpdir = tmp_path_factory.mktemp("aws_l1b_tests")
    return create_mwr_file(tmpdir, fake_mwr_data_array, eps_sterna=False)


@pytest.fixture(scope="module")
def aws_mwr_l1c_file(tmp_path_factory, fake_mwr_data_array):
    """Create an AWS MWR l1c file."""
    tmpdir = tmp_path_factory.mktemp("aws_l1c_tests")
    return create_mwr_file(tmpdir, fake_mwr_data_array, eps_sterna=False, l1b=False)


@pytest.fixture(scope="module")
def eps_sterna_mwr_handler(eps_sterna_mwr_file):
    """Create an EPS-Sterna MWR filehandler."""
    filename_info = parse(file_pattern, os.path.basename(eps_sterna_mwr_file))
    filetype_info = dict()
    filetype_info["file_type"] = "eps_sterna_mwr_l1b"
    filetype_info["feed_horn_group_name"] = "n_feedhorns"
    return AWS_EPS_Sterna_MWR_L1BFile(eps_sterna_mwr_file, filename_info, filetype_info)


@pytest.fixture(scope="module")
def aws_mwr_handler(aws_mwr_file):
    """Create an AWS MWR filehandler."""
    filename_info = parse(file_pattern, os.path.basename(aws_mwr_file))
    filetype_info = dict()
    filetype_info["file_type"] = "aws1_mwr_l1b"
    filetype_info["feed_horn_group_name"] = "n_geo_groups"
    return AWS_EPS_Sterna_MWR_L1BFile(aws_mwr_file, filename_info, filetype_info)


@pytest.fixture(scope="module")
def aws_mwr_l1c_handler(aws_mwr_l1c_file):
    """Create an AWS MWR level-1c filehandler."""
    filename_info = parse(file_pattern, os.path.basename(aws_mwr_l1c_file))
    filetype_info = dict()
    filetype_info["file_type"] = "aws1_mwr_l1c"
    filetype_info["feed_horn_group_name"] = None
    return AWS_MWR_L1CFile(aws_mwr_l1c_file, filename_info, filetype_info)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Satpy developers
#
# This file is part of Satpy.
#
# Satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Module for testing the satpy.readers.oceancolorcci_l3_nc module."""

import datetime as dt
import os

import numpy as np
import pytest
import xarray as xr
from pyresample.geometry import AreaDefinition

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - tmp_path


@pytest.fixture
def fake_dataset():
    """Create a CLAAS-like test dataset."""
    adg = xr.DataArray(
        [[1.0, 0.47, 4.5, 1.2], [0.2, 0, 1.3, 1.3]],
        dims=("lat", "lon")
    )
    atot = xr.DataArray(
        [[0.001, 0.08, 23.4, 0.1], [2.1, 1.2, 4.7, 306.]],
        dims=("lat", "lon")
    )
    kd = xr.DataArray(
        [[0.8, 0.01, 5.34, 1.23], [0.4, 1.0, 3.2, 1.23]],
        dims=("lat", "lon")
    )
    nobs = xr.DataArray(
        [[5, 118, 5, 100], [0, 15, 0, 1]],
        dims=("lat", "lon"),
        attrs={"_FillValue": 0}
    )
    nobs_filt = xr.DataArray(
        [[5, 118, 5, 100], [np.nan, 15, np.nan, 1]],
        dims=("lat", "lon"),
        attrs={"_FillValue": 0}
    )
    watcls = xr.DataArray(
        [[12.2, 0.01, 6.754, 5.33], [12.5, 101.5, 103.5, 204.]],
        dims=("lat", "lon")
    )
    attrs = {
        "geospatial_lon_resolution": "90",
        "geospatial_lat_resolution": "90",
        "geospatial_lon_min": -180.,
        "geospatial_lon_max": 180.,
        "geospatial_lat_min": -90.,
        "geospatial_lat_max": 90.,
        "time_coverage_start": "202108010000Z",
        "time_coverage_end": "202108312359Z",
    }
    return xr.Dataset(
        {
            "adg_490": adg,
            "water_class10": watcls,
            "SeaWiFS_nobs_sum": nobs,
            "test_nobs": nobs_filt,
            "kd_490": kd,
            "atot_665": atot,
        },
        attrs=attrs
    )


ds_dict = {"adg_490": "adg_490",
           "water_class10": "water_class10",
           "seawifs_nobs_sum": "test_nobs",
           "kd_490": "kd_490",
           "atot_665": "atot_665"}

ds_list_all = ["adg_490", "water_class10", "seawifs_nobs_sum", "kd_490", "atot_665"]
ds_list_iop = ["adg_490", "water_class10", "seawifs_nobs_sum", "atot_665"]
ds_list_kd = ["kd_490", "water_class10", "seawifs_nobs_sum"]


@pytest.fixture
def fake_file_dict(fake_dataset, tmp_path):
    """Write a fake dataset to file."""
    fdict = {}
    filename = tmp_path / "ESACCI-OC-L3S-OC_PRODUCTS-MERGED-10M_MONTHLY_4km_GEO_PML_OCx_QAA-202112-fv5.0.nc"
    fake_dataset.to_netcdf(filename)
    fdict["bad_month"] = filename

    filename = tmp_path / "ESACCI-OC-L3S-OC_PRODUCTS-MERGED-2D_DAILY_4km_GEO_PML_OCx_QAA-202112-fv5.0.nc"
    fake_dataset.to_netcdf(filename)
    fdict["bad_day"] = filename

    filename = tmp_path / "ESACCI-OC-L3S-OC_PRODUCTS-MERGED-1M_MONTHLY_4km_GEO_PML_OCx_QAA-202112-fv5.0.nc"
    fake_dataset.to_netcdf(filename)
    fdict["ocprod_1m"] = filename

    filename = tmp_path / "ESACCI-OC-L3S-OC_PRODUCTS-MERGED-5D_DAILY_4km_GEO_PML_OCx_QAA-202112-fv5.0.nc"
    fake_dataset.to_netcdf(filename)
    fdict["ocprod_5d"] = filename

    filename = tmp_path / "ESACCI-OC-L3S-IOP-MERGED-8D_DAILY_4km_GEO_PML_RRS-20211117-fv5.0.nc"
    fake_dataset.to_netcdf(filename)
    fdict["iop_8d"] = filename

    filename = tmp_path / "ESACCI-OC-L3S-IOP-MERGED-1D_DAILY_4km_GEO_PML_OCx-202112-fv5.0.nc"
    fake_dataset.to_netcdf(filename)
    fdict["iop_1d"] = filename

    filename = tmp_path / "ESACCI-OC-L3S-K_490-MERGED-1D_DAILY_4km_GEO_PML_RRS-20210113-fv5.0.nc"
    fake_dataset.to_netcdf(filename)
    fdict["k490_1d"] = filename

    return fdict


class TestOCCCIReader:
    """Test the Ocean Color reader."""

    def setup_method(self):
        """Set up the reader tests."""
        from satpy._config import config_search_paths

        self.yaml_file = "oceancolorcci_l3_nc.yaml"
        self.reader_configs = config_search_paths(os.path.join("readers", self.yaml_file))

    def _create_reader_for_resolutions(self, filename):
        from satpy.readers.core.loading import load_reader
        reader = load_reader(self.reader_configs)
        files = reader.select_files_from_pathnames(filename)
        assert len(filename) == len(files)
        reader.create_filehandlers(files)
        # Make sure we have some files
        assert reader.file_handlers
        return reader

    @pytest.fixture
    def area_exp(self):
        """Get expected area definition."""
        proj_dict = {"datum": "WGS84", "no_defs": "None", "proj": "longlat", "type": "crs"}

        return AreaDefinition(
            area_id="gridded_occci",
            description="Full globe gridded area",
            proj_id="longlat",
            projection=proj_dict,
            area_extent=(-180., -90., 180., 90.),
            width=4,
            height=2,
        )

    def test_get_area_def(self, area_exp, fake_file_dict):
        """Test area definition."""
        import warnings

        reader = self._create_reader_for_resolutions([fake_file_dict["ocprod_1m"]])
        res = reader.load([ds_list_all[0]])
        area = res[ds_list_all[0]].attrs["area"]

        assert area.area_id == area_exp.area_id
        assert area.area_extent == area_exp.area_extent
        assert area.width == area_exp.width
        assert area.height == area_exp.height
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore",
                                    message=r"You will likely lose important projection information",
                                    category=UserWarning)
            # The corresponding CRS objects do not match even if the proj dicts match, so use the dicts
            assert area.proj_dict == area_exp.proj_dict

    def test_bad_fname(self, fake_dataset, fake_file_dict):
        """Test case where an incorrect composite period is given."""
        reader = self._create_reader_for_resolutions([fake_file_dict["bad_month"]])
        res = reader.load([ds_list_all[0]])
        assert len(res) == 0
        reader = self._create_reader_for_resolutions([fake_file_dict["bad_day"]])
        res = reader.load([ds_list_all[0]])
        assert len(res) == 0

    def test_get_dataset_monthly_allprods(self, fake_dataset, fake_file_dict):
        """Test dataset loading."""
        reader = self._create_reader_for_resolutions([fake_file_dict["ocprod_1m"]])
        # Check how many datasets are available. This file contains all of them.
        assert len(list(reader.available_dataset_names)) == 94
        res = reader.load(ds_list_all)
        assert len(res) == len(ds_list_all)
        for curds in ds_list_all:
            np.testing.assert_allclose(res[curds].values, fake_dataset[ds_dict[curds]].values)
            assert res[curds].attrs["sensor"] == "merged"
            assert res[curds].attrs["composite_period"] == "monthly"

    def test_get_dataset_8d_iopprods(self, fake_dataset, fake_file_dict):
        """Test dataset loading."""
        reader = self._create_reader_for_resolutions([fake_file_dict["iop_8d"]])
        # Check how many datasets are available. This file contains all of them.
        assert len(list(reader.available_dataset_names)) == 70
        res = reader.load(ds_list_iop)
        assert len(res) == len(ds_list_iop)
        for curds in ds_list_iop:
            np.testing.assert_allclose(res[curds].values, fake_dataset[ds_dict[curds]].values)
            assert res[curds].attrs["sensor"] == "merged"
            assert res[curds].attrs["composite_period"] == "8-day"

    def test_get_dataset_1d_kprods(self, fake_dataset, fake_file_dict):
        """Test dataset loading."""
        reader = self._create_reader_for_resolutions([fake_file_dict["k490_1d"]])
        # Check how many datasets are available. This file contains all of them.
        assert len(list(reader.available_dataset_names)) == 25
        res = reader.load(ds_list_kd)
        assert len(res) == len(ds_list_kd)
        for curds in ds_list_kd:
            np.testing.assert_allclose(res[curds].values, fake_dataset[ds_dict[curds]].values)
            assert res[curds].attrs["sensor"] == "merged"
            assert res[curds].attrs["composite_period"] == "daily"

    def test_get_dataset_5d_allprods(self, fake_dataset, fake_file_dict):
        """Test dataset loading."""
        reader = self._create_reader_for_resolutions([fake_file_dict["ocprod_5d"]])
        # Check how many datasets are available. This file contains all of them.
        assert len(list(reader.available_dataset_names)) == 94
        res = reader.load(ds_list_all)
        assert len(res) == len(ds_list_all)
        for curds in ds_list_all:
            np.testing.assert_allclose(res[curds].values, fake_dataset[ds_dict[curds]].values)
            assert res[curds].attrs["sensor"] == "merged"
            assert res[curds].attrs["composite_period"] == "5-day"

    def test_start_time(self, fake_file_dict):
        """Test start time property."""
        reader = self._create_reader_for_resolutions([fake_file_dict["k490_1d"]])
        assert reader.start_time == dt.datetime(2021, 8, 1, 0, 0, 0)

    def test_end_time(self, fake_file_dict):
        """Test end time property."""
        reader = self._create_reader_for_resolutions([fake_file_dict["iop_8d"]])
        assert reader.end_time == dt.datetime(2021, 8, 31, 23, 59, 0)

    def test_correct_dimnames(self, fake_file_dict):
        """Check that the loaded dimension names are correct."""
        reader = self._create_reader_for_resolutions([fake_file_dict["ocprod_5d"]])
        res = reader.load(ds_list_all)
        for dsname in ds_list_all:
            assert res[dsname].dims[0] == "y"
            assert res[dsname].dims[1] == "x"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Common functionality for FCI data readers."""
from __future__ import annotations

# Platform names according to the MTG FCI L1 Product User Guide,
# EUM/MTG/USR/13/719113 from 2019-06-27, pages 32 and 124, are MTI1, MTI2,
# MTI3, and MTI4, but we want to use names such as described in WMO OSCAR
# MTG-I1, MTG-I2, MTG-I3, and MTG-I4.
#
# Not sure how the numbering will be considering MTG-S1 and MTG-S2 will be launched
# in-between.
platform_name_translate = {
    "MTI1": "Meteosat-12",
    "MTI2": "MTG-I2",
    "MTI3": "MTG-I3",
    "MTI4": "MTG-I4"}


def calculate_area_extent(area_dict):
        """Calculate the area extent seen by MTG FCI instrument.

        Since the center of the FCI grids is located at the interface between the pixels, there are equally many
        pixels (e.g. 5568/2 = 2784 for 2km grid) in each direction from the center points. Hence, the area extent
        can be easily computed by simply adding and subtracting half the width and height from teh centre point (=0).

        Args:
            area_dict: A dictionary containing the required parameters
                ncols: number of pixels in east-west direction
                nlines: number of pixels in south-north direction
                column_step: Pixel resulution in meters in east-west direction
                line_step: Pixel resulution in meters in south-north direction
        Returns:
            tuple: An area extent for the scene defined by the lower left and
                   upper right corners

        """
        ncols = area_dict["ncols"]
        nlines = area_dict["nlines"]
        column_step = area_dict["column_step"]
        line_step = area_dict["line_step"]

        ll_c = (0 - ncols / 2.) * column_step
        ll_l = (0 + nlines / 2.) * line_step
        ur_c = (0 + ncols / 2.) * column_step
        ur_l = (0 - nlines / 2.) * line_step

        return (ll_c, ll_l, ur_c, ur_l)



#!/usr/bin/env python
# Copyright (c) 2018-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for sharpening compositors."""

import datetime as dt
import unittest
from unittest import mock

import dask.array as da
import numpy as np
import pytest
import xarray as xr

from satpy.tests.utils import RANDOM_GEN


class TestRatioSharpenedCompositors:
    """Test RatioSharpenedRGB and SelfSharpendRGB compositors."""

    def setup_method(self):
        """Create test data."""
        from pyresample.geometry import AreaDefinition
        area = AreaDefinition("test", "test", "test",
                              {"proj": "merc"}, 2, 2,
                              (-2000, -2000, 2000, 2000))
        attrs = {"area": area,
                 "start_time": dt.datetime(2018, 1, 1, 18),
                 "modifiers": tuple(),
                 "resolution": 1000,
                 "calibration": "reflectance",
                 "units": "%",
                 "name": "test_vis"}
        low_res_data = np.ones((2, 2), dtype=np.float64) + 4
        low_res_data[1, 1] = 0.0  # produces infinite ratio
        ds1 = xr.DataArray(da.from_array(low_res_data, chunks=2),
                           attrs=attrs, dims=("y", "x"),
                           coords={"y": [0, 1], "x": [0, 1]})
        self.ds1 = ds1

        ds2 = xr.DataArray(da.ones((2, 2), chunks=2, dtype=np.float64) + 2,
                           attrs=attrs, dims=("y", "x"),
                           coords={"y": [0, 1], "x": [0, 1]})
        ds2.attrs["name"] += "2"
        self.ds2 = ds2

        ds3 = xr.DataArray(da.ones((2, 2), chunks=2, dtype=np.float64) + 3,
                           attrs=attrs, dims=("y", "x"),
                           coords={"y": [0, 1], "x": [0, 1]})
        ds3.attrs["name"] += "3"
        self.ds3 = ds3

        # high resolution version
        high_res_data = np.ones((2, 2), dtype=np.float64)
        high_res_data[1, 0] = np.nan  # invalid value in one band
        ds4 = xr.DataArray(da.from_array(high_res_data, chunks=2),
                           attrs=attrs, dims=("y", "x"),
                           coords={"y": [0, 1], "x": [0, 1]})
        ds4.attrs["name"] += "4"
        ds4.attrs["resolution"] = 500
        self.ds4 = ds4

        # high resolution version - but too big
        ds4_big = xr.DataArray(da.ones((4, 4), chunks=2, dtype=np.float64),
                               attrs=attrs.copy(), dims=("y", "x"),
                               coords={"y": [0, 1, 2, 3], "x": [0, 1, 2, 3]})
        ds4_big.attrs["name"] += "4"
        ds4_big.attrs["resolution"] = 500
        ds4_big.attrs["rows_per_scan"] = 1
        ds4_big.attrs["area"] = AreaDefinition("test", "test", "test",
                                               {"proj": "merc"}, 4, 4,
                                               (-2000, -2000, 2000, 2000))
        self.ds4_big = ds4_big

    @pytest.mark.parametrize(
        "init_kwargs",
        [
            {"high_resolution_band": "bad", "neutral_resolution_band": "red"},
            {"high_resolution_band": "red", "neutral_resolution_band": "bad"}
        ]
    )
    def test_bad_colors(self, init_kwargs):
        """Test that only valid band colors can be provided."""
        from satpy.composites.resolution import RatioSharpenedRGB
        with pytest.raises(ValueError, match="RatioSharpenedRGB..*_band must be one of .*"):
            RatioSharpenedRGB(name="true_color", **init_kwargs)

    def test_match_data_arrays(self):
        """Test that all areas have to be the same resolution."""
        from satpy.composites.resolution import IncompatibleAreas, RatioSharpenedRGB
        comp = RatioSharpenedRGB(name="true_color")
        with pytest.raises(IncompatibleAreas):
            comp((self.ds1, self.ds2, self.ds3), optional_datasets=(self.ds4_big,))

    def test_more_than_three_datasets(self):
        """Test that only 3 datasets can be passed."""
        from satpy.composites.resolution import RatioSharpenedRGB
        comp = RatioSharpenedRGB(name="true_color")
        with pytest.raises(ValueError, match="Expected 3 datasets, got 4"):
            comp((self.ds1, self.ds2, self.ds3, self.ds1), optional_datasets=(self.ds4_big,))

    def test_self_sharpened_no_high_res(self):
        """Test for exception when no high_res band is specified."""
        from satpy.composites.resolution import SelfSharpenedRGB
        comp = SelfSharpenedRGB(name="true_color", high_resolution_band=None)
        with pytest.raises(ValueError, match="SelfSharpenedRGB requires at least one high resolution band, not 'None'"):
            comp((self.ds1, self.ds2, self.ds3))

    @pytest.mark.parametrize("dtype", [np.float32, np.float64])
    def test_basic_no_high_res(self, dtype):
        """Test that three datasets can be passed without optional high res."""
        from satpy.composites.resolution import RatioSharpenedRGB
        comp = RatioSharpenedRGB(name="true_color")
        res = comp((self.ds1.astype(dtype), self.ds2.astype(dtype), self.ds3.astype(dtype)))
        assert res.shape == (3, 2, 2)
        assert res.dtype == dtype
        assert res.values.dtype == dtype

    @pytest.mark.parametrize("dtype", [np.float32, np.float64])
    def test_basic_no_sharpen(self, dtype):
        """Test that color None does no sharpening."""
        from satpy.composites.resolution import RatioSharpenedRGB
        comp = RatioSharpenedRGB(name="true_color", high_resolution_band=None)
        res = comp((self.ds1.astype(dtype), self.ds2.astype(dtype), self.ds3.astype(dtype)),
                   optional_datasets=(self.ds4.astype(dtype),))
        assert res.shape == (3, 2, 2)
        assert res.dtype == dtype
        assert res.values.dtype == dtype

    @pytest.mark.parametrize("dtype", [np.float32, np.float64])
    @pytest.mark.parametrize(
        ("high_resolution_band", "neutral_resolution_band", "exp_r", "exp_g", "exp_b"),
        [
            ("red", None,
             np.array([[1.0, 1.0], [np.nan, 1.0]], dtype=np.float64),
             np.array([[0.6, 0.6], [np.nan, 3.0]], dtype=np.float64),
             np.array([[0.8, 0.8], [np.nan, 4.0]], dtype=np.float64)),
            ("red", "green",
             np.array([[1.0, 1.0], [np.nan, 1.0]], dtype=np.float64),
             np.array([[3.0, 3.0], [np.nan, 3.0]], dtype=np.float64),
             np.array([[0.8, 0.8], [np.nan, 4.0]], dtype=np.float64)),
            ("green", None,
             np.array([[5 / 3, 5 / 3], [np.nan, 0.0]], dtype=np.float64),
             np.array([[1.0, 1.0], [np.nan, 1.0]], dtype=np.float64),
             np.array([[4 / 3, 4 / 3], [np.nan, 4 / 3]], dtype=np.float64)),
            ("green", "blue",
             np.array([[5 / 3, 5 / 3], [np.nan, 0.0]], dtype=np.float64),
             np.array([[1.0, 1.0], [np.nan, 1.0]], dtype=np.float64),
             np.array([[4.0, 4.0], [np.nan, 4.0]], dtype=np.float64)),
            ("blue", None,
             np.array([[1.25, 1.25], [np.nan, 0.0]], dtype=np.float64),
             np.array([[0.75, 0.75], [np.nan, 0.75]], dtype=np.float64),
             np.array([[1.0, 1.0], [np.nan, 1.0]], dtype=np.float64)),
            ("blue", "red",
             np.array([[5.0, 5.0], [np.nan, 0.0]], dtype=np.float64),
             np.array([[0.75, 0.75], [np.nan, 0.75]], dtype=np.float64),
             np.array([[1.0, 1.0], [np.nan, 1.0]], dtype=np.float64))
        ]
    )
    def test_ratio_sharpening(self, high_resolution_band, neutral_resolution_band, exp_r, exp_g, exp_b, dtype):
        """Test RatioSharpenedRGB by different groups of high_resolution_band and neutral_resolution_band."""
        from satpy.composites.resolution import RatioSharpenedRGB
        comp = RatioSharpenedRGB(name="true_color", high_resolution_band=high_resolution_band,
                                 neutral_resolution_band=neutral_resolution_band)
        res = comp((self.ds1.astype(dtype), self.ds2.astype(dtype), self.ds3.astype(dtype)),
                   optional_datasets=(self.ds4.astype(dtype),))

        assert "units" not in res.attrs
        assert isinstance(res, xr.DataArray)
        assert isinstance(res.data, da.Array)
        assert res.dtype == dtype

        data = res.values
        np.testing.assert_allclose(data[0], exp_r, rtol=1e-5)
        np.testing.assert_allclose(data[1], exp_g, rtol=1e-5)
        np.testing.assert_allclose(data[2], exp_b, rtol=1e-5)
        assert res.dtype == dtype

    @pytest.mark.parametrize("dtype", [np.float32, np.float64])
    @pytest.mark.parametrize(
        ("exp_shape", "exp_r", "exp_g", "exp_b"),
        [
            ((3, 2, 2),
             np.array([[5.0, 5.0], [5.0, 0]], dtype=np.float64),
             np.array([[4.0, 4.0], [4.0, 0]], dtype=np.float64),
             np.array([[16 / 3, 16 / 3], [16 / 3, 0]], dtype=np.float64))
        ]
    )
    def test_self_sharpened_basic(self, exp_shape, exp_r, exp_g, exp_b, dtype):
        """Test that three datasets can be passed without optional high res."""
        from satpy.composites.resolution import SelfSharpenedRGB
        comp = SelfSharpenedRGB(name="true_color")
        res = comp((self.ds1.astype(dtype), self.ds2.astype(dtype), self.ds3.astype(dtype)))
        assert res.dtype == dtype

        data = res.values
        assert data.shape == exp_shape
        np.testing.assert_allclose(data[0], exp_r, rtol=1e-5)
        np.testing.assert_allclose(data[1], exp_g, rtol=1e-5)
        np.testing.assert_allclose(data[2], exp_b, rtol=1e-5)
        assert data.dtype == dtype


class TestLuminanceSharpeningCompositor(unittest.TestCase):
    """Test luminance sharpening compositor."""

    def test_compositor(self):
        """Test luminance sharpening compositor."""
        from satpy.composites.resolution import LuminanceSharpeningCompositor
        comp = LuminanceSharpeningCompositor(name="test")
        # Three shades of grey
        rgb_arr = np.array([1, 50, 100, 200, 1, 50, 100, 200, 1, 50, 100, 200])
        rgb = xr.DataArray(rgb_arr.reshape((3, 2, 2)),
                           dims=["bands", "y", "x"], coords={"bands": ["R", "G", "B"]})
        # 100 % luminance -> all result values ~1.0
        lum = xr.DataArray(np.array([[100., 100.], [100., 100.]]),
                           dims=["y", "x"])
        res = comp([lum, rgb])
        np.testing.assert_allclose(res.data, 1., atol=1e-9)
        # 50 % luminance, all result values ~0.5
        lum = xr.DataArray(np.array([[50., 50.], [50., 50.]]),
                           dims=["y", "x"])
        res = comp([lum, rgb])
        np.testing.assert_allclose(res.data, 0.5, atol=1e-9)
        # 30 % luminance, all result values ~0.3
        lum = xr.DataArray(np.array([[30., 30.], [30., 30.]]),
                           dims=["y", "x"])
        res = comp([lum, rgb])
        np.testing.assert_allclose(res.data, 0.3, atol=1e-9)
        # 0 % luminance, all values ~0.0
        lum = xr.DataArray(np.array([[0., 0.], [0., 0.]]),
                           dims=["y", "x"])
        res = comp([lum, rgb])
        np.testing.assert_allclose(res.data, 0.0, atol=1e-9)


class TestSandwichCompositor:
    """Test sandwich compositor."""

    # Test RGB and RGBA
    @pytest.mark.parametrize(
        ("input_shape", "bands"),
        [
            ((3, 2, 2), ["R", "G", "B"]),
            ((4, 2, 2), ["R", "G", "B", "A"])
        ]
    )
    @mock.patch("satpy.composites.resolution.enhance2dataset")
    def test_compositor(self, e2d, input_shape, bands):
        """Test luminance sharpening compositor."""
        from satpy.composites.resolution import SandwichCompositor

        rgb_arr = da.from_array(RANDOM_GEN.random(input_shape), chunks=2)
        rgb = xr.DataArray(rgb_arr, dims=["bands", "y", "x"],
                           coords={"bands": bands})
        lum_arr = da.from_array(100 * RANDOM_GEN.random((2, 2)), chunks=2)
        lum = xr.DataArray(lum_arr, dims=["y", "x"])

        # Make enhance2dataset return unmodified dataset
        e2d.return_value = rgb
        comp = SandwichCompositor(name="test")

        res = comp([lum, rgb])

        for band in rgb:
            if band.bands != "A":
                # Check compositor has modified this band
                np.testing.assert_allclose(res.loc[band.bands].to_numpy(),
                                           band.to_numpy() * lum_arr / 100.)
            else:
                # Check Alpha band remains intact
                np.testing.assert_allclose(res.loc[band.bands].to_numpy(),
                                           band.to_numpy())
        # make sure the compositor doesn't modify the input data
        np.testing.assert_allclose(lum.values, lum_arr.compute())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2022 Satpy Developers

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Passive Microwave instrument and channel specific features."""

import numbers
from contextlib import suppress
from typing import NamedTuple

import numpy as np


class FrequencyBandBaseArithmetics:
    """Mixin class with basic frequency comparison operations."""

    def __lt__(self, other):
        """Compare to another frequency."""
        if other is None:
            return False
        return super().__lt__(other)

    def __gt__(self, other):
        """Compare to another frequency."""
        if other is None:
            return True
        return super().__gt__(other)

    @classmethod
    def convert(cls, frq):
        """Convert `frq` to this type if possible."""
        if isinstance(frq, dict):
            return cls(**frq)
        return frq


class FrequencyQuadrupleSideBandBase(NamedTuple):
    """Base class for a frequency quadruple side band.

    Frequency Quadruple Side Band is supposed to describe the special type of
    bands commonly used in temperature sounding from Passive Microwave
    Sensors. When the absorption band being observed is symmetrical it is
    advantageous (giving better NeDT) to sense in a band both right and left of
    the central absorption frequency. But to avoid (CO2) absorption lines
    symmetrically positioned on each side of the main absorption band it is
    common to split the side bands in two 'side-side' bands.

    This is needed because of this bug: https://bugs.python.org/issue41629

    """

    central: float
    side: float
    sideside: float
    bandwidth: float
    unit: str = "GHz"


class FrequencyQuadrupleSideBand(FrequencyBandBaseArithmetics, FrequencyQuadrupleSideBandBase):
    """The frequency quadruple side band class.

    The elements of the quadruple-side-band type frequency band are the
    central frquency, the relative (main) side band frequency (relative to the
    center - left and right), the sub-side band frequency (relative to the
    offset side-band(s)) and their bandwidths. Optionally a unit (defaults to
    GHz) may be specified. No clever unit conversion is done here, it's just
    used for checking that two ranges are comparable.

    Frequency Quadruple Side Band is supposed to describe the special type of
    bands commonly used in temperature sounding from Passive Microwave
    Sensors. When the absorption band being observed is symmetrical it is
    advantageous (giving better NeDT) to sense in a band both right and left of
    the central absorption frequency. But to avoid (CO2) absorption lines
    symmetrically positioned on each side of the main absorption band it is
    common to split the side bands in two 'side-side' bands.

    """

    def __eq__(self, other):
        """Return if two channel frequencies are equal.

        Args:
            other (tuple[numbers.Number] or numbers.Number): (central frq, side band frq, side-side band frq,
                and band width frq) or scalar frq

        Return:
            (bool): True if other is a scalar and min <= other <= max, or if other is a
                tuple equal to self, or if other is a number contained by self.
                False otherwise.

        """
        if other is None:
            return False
        if isinstance(other, numbers.Number):
            return other in self
        if isinstance(other, (tuple, list)) and len(other) == 4:
            return other in self
        return super().__eq__(other)

    def __str__(self):
        """Format for print out."""
        return f"central={self.central} {self.unit} ±{self.side} ±{self.sideside} width={self.bandwidth} {self.unit}"

    def __hash__(self):
        """Hash this tuple."""
        return tuple.__hash__(self)

    def __contains__(self, other):
        """Check if this quadruple-side-band 'contains' *other*."""
        if other is None:
            return False

        # The four centrals:
        central_left_left = self.central - self.side - self.sideside
        central_left_right = self.central - self.side + self.sideside
        central_right_left = self.central + self.side - self.sideside
        central_right_right = self.central + self.side + self.sideside

        four_centrals = [central_left_left, central_left_right,
                         central_right_left, central_right_right]
        if isinstance(other, numbers.Number):
            for central in four_centrals:
                if _is_inside_interval(other, central, self.bandwidth):
                    return True

            return False

        if isinstance(other, (tuple, list)) and len(other) == 5:
            raise NotImplementedError("Can't check if one frequency quadruple side band is contained in another.")

        with suppress(AttributeError):
            if self.unit != other.unit:
                raise NotImplementedError("Can't compare frequency ranges with different units.")

        return False

    def distance(self, value):
        """Get the distance to the quadruple side band.

        Determining the distance in frequency space between two quadruple side
        bands can be quite ambiguous, as such bands are in effect a set of 4
        narrow bands, two on each side of the main absorption band, and on each
        side, one on each side of the secondary absorption lines. To keep it as
        simple as possible we have until further decided to define the distance
        between such two bands to infinity if they are determined to be equal.

        If the frequency entered is a single value, the distance will be the
        minimum of the distances to the two outermost sides of the quadruple
        side band.

        If the frequency entered is a tuple or list and the two quadruple
        frequency bands are contained in each other (equal) the distance will
        always be zero.

        """
        left_left = self.central - self.side - self.sideside
        right_right = self.central + self.side + self.sideside

        if self == value:
            try:
                left_side_dist = abs(value.central - value.side - value.sideside - left_left)
                right_side_dist = abs(value.central + value.side + value.sideside - right_right)
            except AttributeError:
                left_side_dist = abs(value - left_left)
                right_side_dist = abs(value - right_right)

            return min(left_side_dist, right_side_dist)
        else:
            return np.inf


class FrequencyDoubleSideBandBase(NamedTuple):
    """Base class for a frequency double side band.

    Frequency Double Side Band is supposed to describe the special type of bands
    commonly used in humidty sounding from Passive Microwave Sensors. When the
    absorption band being observed is symmetrical it is advantageous (giving
    better NeDT) to sense in a band both right and left of the central
    absorption frequency.

    This is needed because of this bug: https://bugs.python.org/issue41629

    """

    central: float
    side: float
    bandwidth: float
    unit: str = "GHz"


class FrequencyDoubleSideBand(FrequencyBandBaseArithmetics, FrequencyDoubleSideBandBase):
    """The frequency double side band class.

    The elements of the double-side-band type frequency band are the central
    frquency, the relative side band frequency (relative to the center - left
    and right) and their bandwidths, and optionally a unit (defaults to
    GHz). No clever unit conversion is done here, it's just used for checking
    that two ranges are comparable.

    Frequency Double Side Band is supposed to describe the special type of bands
    commonly used in humidty sounding from Passive Microwave Sensors. When the
    absorption band being observed is symmetrical it is advantageous (giving
    better NeDT) to sense in a band both right and left of the central
    absorption frequency.

    """

    def __eq__(self, other):
        """Return if two channel frequencies are equal.

        Args:
            other (tuple[numbers.Number] or numbers.Number): (central frq, side band frq and band width frq)
                or scalar frq

        Return:
            (bool): True if other is a scalar and min <= other <= max, or if other is a
                tuple equal to self, or if other is a number contained by self.
                False otherwise.

        """
        if other is None:
            return False
        if isinstance(other, numbers.Number):
            return other in self
        if isinstance(other, (tuple, list)) and len(other) == 3:
            return other in self
        return super().__eq__(other)

    def __str__(self):
        """Format for print out."""
        return f"central={self.central} {self.unit} ±{self.side} width={self.bandwidth} {self.unit}"

    def __hash__(self):
        """Hash this tuple."""
        return tuple.__hash__(self)

    def __contains__(self, other):
        """Check if this double-side-band 'contains' *other*."""
        if other is None:
            return False

        leftside = self.central - self.side
        rightside = self.central + self.side

        if isinstance(other, numbers.Number):
            if self._check_band_contains_other((leftside, self.bandwidth), (other, 0)):
                return True
            return self._check_band_contains_other((rightside, self.bandwidth), (other, 0))

        other_leftside, other_rightside, other_bandwidth = 0, 0, 0
        if isinstance(other, (tuple, list)) and len(other) == 3:
            other_leftside = other[0] - other[1]
            other_rightside = other[0] + other[1]
            other_bandwidth = other[2]
        else:
            with suppress(AttributeError):
                if self.unit != other.unit:
                    raise NotImplementedError("Can't compare frequency ranges with different units.")
                other_leftside = other.central - other.side
                other_rightside = other.central + other.side
                other_bandwidth = other.bandwidth

        if self._check_band_contains_other((leftside, self.bandwidth), (other_leftside, other_bandwidth)):
            return True
        return self._check_band_contains_other((rightside, self.bandwidth), (other_rightside, other_bandwidth))

    @staticmethod
    def _check_band_contains_other(band, other_band):
        """Check that a band contains another band.

        A band is here defined as a tuple of a central frequency and a bandwidth.
        """
        central1, width1 = band
        central_other, width_other = other_band

        if ((central1 - width1/2. <= central_other - width_other/2.) and
                (central1 + width1/2. >= central_other + width_other/2.)):
            return True
        return False

    def distance(self, value):
        """Get the distance to the double side band.

        Determining the distance in frequency space between two double side
        bands can be quite ambiguous, as such bands are in effect a set of 2
        narrow bands, one on each side of the absorption line. To keep it
        as simple as possible we have until further decided to set the
        distance between such two bands to infitiy if neither of them are
        contained in the other.

        If the frequency entered is a single value and this frequency falls
        inside one of the side bands, the distance will be the minimum of the
        distances to the two outermost sides of the double side band. However,
        is such a single frequency value falls outside one of the two side
        bands, the distance will be set to infitiy.

        If the frequency entered is a tuple the distance will either be 0 (if
        one is containde in the other) or infinity.
        """
        if self == value:
            try:
                left_side_dist = abs(value.central - value.side - (self.central - self.side))
                right_side_dist = abs(value.central + value.side - (self.central + self.side))
            except AttributeError:
                if isinstance(value, (tuple, list)):
                    return abs((value[0] - value[1]) - (self.central - self.side))

                left_side_dist = abs(value - (self.central - self.side))
                right_side_dist = abs(value - (self.central + self.side))

            return min(left_side_dist, right_side_dist)
        else:
            return np.inf


class FrequencyRangeBase(NamedTuple):
    """Base class for frequency ranges.

    This is needed because of this bug: https://bugs.python.org/issue41629
    """

    central: float
    bandwidth: float
    unit: str = "GHz"


class FrequencyRange(FrequencyBandBaseArithmetics, FrequencyRangeBase):
    """The Frequency range class.

    The elements of the range are central and bandwidth values, and optionally
    a unit (defaults to GHz). No clever unit conversion is done here, it's just
    used for checking that two ranges are comparable.

    This type is used for passive microwave sensors.

    """

    def __eq__(self, other):
        """Check wether two channel frequencies are equal.

        Args:
            other (tuple[numbers.Number] or numbers.Number): (central frq, band width frq) or scalar frq

        Return:
            (bool): True if other is a scalar and min <= other <= max, or if other is a
                tuple equal to self, or if other is a number contained by self.
                False otherwise.

        """
        if other is None:
            return False
        if isinstance(other, numbers.Number):
            return other in self
        if isinstance(other, (tuple, list)) and len(other) == 2:
            return self[:2] == other
        return super().__eq__(other)

    def __str__(self):
        """Format for print out."""
        return f"central={self.central} {self.unit} width={self.bandwidth} {self.unit}"

    def __hash__(self):
        """Hash this tuple."""
        return tuple.__hash__(self)

    def __contains__(self, other):
        """Check if this range contains *other*."""
        if other is None:
            return False
        if isinstance(other, numbers.Number):
            return self.central - self.bandwidth/2. <= other <= self.central + self.bandwidth/2.

        with suppress(AttributeError):
            if self.unit != other.unit:
                raise NotImplementedError("Can't compare frequency ranges with different units.")
            return (self.central - self.bandwidth/2. <= other.central - other.bandwidth/2. and
                    self.central + self.bandwidth/2. >= other.central + other.bandwidth/2.)
        return False

    def distance(self, value):
        """Get the distance from value."""
        if self == value:
            try:
                return abs(value.central - self.central)
            except AttributeError:
                if isinstance(value, (tuple, list)):
                    return abs(value[0] - self.central)
                return abs(value - self.central)
        else:
            return np.inf


def _is_inside_interval(value, central, width):
    return central - width/2 <= value <= central + width/2



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2016-2020 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""SLSTR L1b reader."""

import datetime as dt
import logging
import os
import re
import warnings

import dask.array as da
import numpy as np
import xarray as xr

from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.utils import get_legacy_chunk_size

logger = logging.getLogger(__name__)

CHUNK_SIZE = get_legacy_chunk_size()

PLATFORM_NAMES = {"S3A": "Sentinel-3A",
                  "S3B": "Sentinel-3B",
                  "S3C": "Sentinel-3C",
                  "S3D": "Sentinel-3D"}

# These are the default channel adjustment factors.
# Defined in the product notice: S3.PN-SLSTR-L1.10
# https://user.eumetsat.int/s3/eup-strapi-media/S3_PN_SLSTR_L1_10_i1r0_SLSTR_L1_PB_SL_L1_004_05_00_fbf01b8813.pdf
CHANCALIB_FACTORS = {"S1_nadir": 0.97,
                     "S2_nadir": 0.98,
                     "S3_nadir": 0.98,
                     "S4_nadir": 1.0,
                     "S5_nadir": 1.11,
                     "S6_nadir": 1.13,
                     "S7_nadir": 1.0,
                     "S8_nadir": 1.0,
                     "S9_nadir": 1.0,
                     "S1_oblique": 0.94,
                     "S2_oblique": 0.95,
                     "S3_oblique": 0.95,
                     "S4_oblique": 1.0,
                     "S5_oblique": 1.04,
                     "S6_oblique": 1.07,
                     "S7_oblique": 1.0,
                     "S8_oblique": 1.0,
                     "S9_oblique": 1.0, }


class NCSLSTRGeo(BaseFileHandler):
    """Filehandler for geo info."""

    def __init__(self, filename, filename_info, filetype_info):
        """Initialize the geo filehandler."""
        super(NCSLSTRGeo, self).__init__(filename, filename_info,
                                         filetype_info)
        self.nc = xr.open_dataset(self.filename,
                                  decode_cf=True,
                                  mask_and_scale=True,
                                  chunks={"columns": CHUNK_SIZE,
                                          "rows": CHUNK_SIZE})
        self.nc = self.nc.rename({"columns": "x", "rows": "y"})

        self.cache = {}

    def get_dataset(self, key, info):
        """Load a dataset."""
        logger.debug("Reading %s.", key["name"])
        file_key = info["file_key"].format(view=key["view"].name[0],
                                           stripe=key["stripe"].name)
        try:
            variable = self.nc[file_key]
        except KeyError:
            return

        info = info.copy()
        info.update(variable.attrs)

        variable.attrs = info
        return variable

    @property
    def start_time(self):
        """Get the start time."""
        return dt.datetime.strptime(self.nc.attrs["start_time"], "%Y-%m-%dT%H:%M:%S.%fZ")

    @property
    def end_time(self):
        """Get the end time."""
        return dt.datetime.strptime(self.nc.attrs["stop_time"], "%Y-%m-%dT%H:%M:%S.%fZ")


class NCSLSTR1B(BaseFileHandler):
    """Filehandler for l1 SLSTR data.

    By default, the calibration factors recommended by EUMETSAT are applied.
    This is required as the SLSTR VIS channels are producing slightly incorrect
    radiances that require adjustment.
    Satpy uses the radiance corrections in S3.PN-SLSTR-L1.08, checked 11/03/2022.
    User-supplied coefficients can be passed via the `user_calibration` kwarg
    This should be a dict of channel names (such as `S1_nadir`, `S8_oblique`).

    For example::

        calib_dict = {'S1_nadir': 1.12}
        scene = satpy.Scene(filenames,
                            reader='slstr-l1b',
                            reader_kwargs={'user_calib': calib_dict})

    Will multiply S1 nadir radiances by 1.12.
    """

    def __init__(self, filename, filename_info, filetype_info,
                 user_calibration=None):
        """Initialize the SLSTR l1 data filehandler."""
        super(NCSLSTR1B, self).__init__(filename, filename_info,
                                        filetype_info)

        self.nc = xr.open_dataset(self.filename,
                                  decode_cf=True,
                                  mask_and_scale=True,
                                  chunks={"columns": CHUNK_SIZE,
                                          "rows": CHUNK_SIZE})
        self.nc = self.nc.rename({"columns": "x", "rows": "y"})
        self.channel = filename_info["dataset_name"]
        self.stripe = filename_info["stripe"]
        views = {"n": "nadir", "o": "oblique"}
        self.view = views[filename_info["view"]]
        cal_file = os.path.join(os.path.dirname(self.filename), "viscal.nc")
        self.cal = xr.open_dataset(cal_file,
                                   decode_cf=True,
                                   mask_and_scale=True,
                                   chunks={"views": CHUNK_SIZE})
        indices_file = os.path.join(os.path.dirname(self.filename),
                                    "indices_{}{}.nc".format(self.stripe, self.view[0]))
        self.indices = xr.open_dataset(indices_file,
                                       decode_cf=True,
                                       mask_and_scale=True,
                                       chunks={"columns": CHUNK_SIZE,
                                               "rows": CHUNK_SIZE})
        self.indices = self.indices.rename({"columns": "x", "rows": "y"})

        self.platform_name = PLATFORM_NAMES[filename_info["mission_id"]]
        self.sensor = "slstr"
        if isinstance(user_calibration, dict):
            self.usercalib = user_calibration
        else:
            self.usercalib = None

    def _apply_radiance_adjustment(self, radiances):
        """Adjust SLSTR radiances with default or user supplied values."""
        chan_name = self.channel + "_" + self.view
        adjust_fac = None
        if self.usercalib is not None:
            # If user supplied adjustment, use it.
            if chan_name in self.usercalib:
                adjust_fac = self.usercalib[chan_name]
        if adjust_fac is None:
            if chan_name in CHANCALIB_FACTORS:
                adjust_fac = CHANCALIB_FACTORS[chan_name]
            else:
                warnings.warn(
                    "Warning: No radiance adjustment supplied " +
                    "for channel " + chan_name,
                    stacklevel=3
                )
                return radiances
        return radiances * adjust_fac

    @staticmethod
    def _cal_rad(rad, didx, solar_flux=None):
        """Calibrate."""
        indices = np.isfinite(didx)
        rad[indices] /= solar_flux[didx[indices].astype(int)]
        return rad

    def get_dataset(self, key, info):
        """Load a dataset."""
        if (self.channel not in key["name"] or
                self.stripe != key["stripe"].name or
                self.view != key["view"].name):
            return
        logger.debug("Reading %s.", key["name"])
        if key["calibration"] == "brightness_temperature":
            variable = self.nc["{}_BT_{}{}".format(self.channel, self.stripe, self.view[0])]
        else:
            variable = self.nc["{}_radiance_{}{}".format(self.channel, self.stripe, self.view[0])]
        radiances = self._apply_radiance_adjustment(variable)
        units = variable.attrs["units"]
        if key["calibration"] == "reflectance":
            # TODO take into account sun-earth distance
            solar_flux = self.cal[re.sub("_[^_]*$", "", key["name"]) + "_solar_irradiances"]
            d_index = self.indices["detector_{}{}".format(self.stripe, self.view[0])]
            idx = 0 if self.view[0] == "n" else 1  # 0: Nadir view, 1: oblique (check).
            radiances.data = da.map_blocks(
                self._cal_rad, radiances.data, d_index.data, solar_flux=solar_flux[:, idx].values)
            radiances *= np.pi * 100
            units = "%"

        info = info.copy()
        info.update(radiances.attrs)
        info.update(key.to_dict())
        info.update(dict(units=units,
                         platform_name=self.platform_name,
                         sensor=self.sensor,
                         view=self.view))

        radiances.attrs = info
        return radiances

    @property
    def start_time(self):
        """Get the start time."""
        return dt.datetime.strptime(self.nc.attrs["start_time"], "%Y-%m-%dT%H:%M:%S.%fZ")

    @property
    def end_time(self):
        """Get the end time."""
        return dt.datetime.strptime(self.nc.attrs["stop_time"], "%Y-%m-%dT%H:%M:%S.%fZ")


class NCSLSTRAngles(BaseFileHandler):
    """Filehandler for angles."""

    def _loadcart(self, fname):
        """Load a cartesian file of appropriate type."""
        cartf = xr.open_dataset(fname,
                                decode_cf=True,
                                mask_and_scale=True,
                                chunks={"columns": CHUNK_SIZE,
                                        "rows": CHUNK_SIZE})
        return cartf

    def __init__(self, filename, filename_info, filetype_info):
        """Initialize the angles reader."""
        super(NCSLSTRAngles, self).__init__(filename, filename_info,
                                            filetype_info)

        self.nc = xr.open_dataset(self.filename,
                                  decode_cf=True,
                                  mask_and_scale=True,
                                  chunks={"columns": CHUNK_SIZE,
                                          "rows": CHUNK_SIZE})

        # TODO: get metadata from the manifest file (xfdumanifest.xml)
        self.platform_name = PLATFORM_NAMES[filename_info["mission_id"]]
        self.sensor = "slstr"
        self.view = filename_info["view"]
        self._start_time = filename_info["start_time"]
        self._end_time = filename_info["end_time"]

        carta_file = os.path.join(
            os.path.dirname(self.filename), "cartesian_a{}.nc".format(self.view[0]))
        carti_file = os.path.join(
            os.path.dirname(self.filename), "cartesian_i{}.nc".format(self.view[0]))
        cartx_file = os.path.join(
            os.path.dirname(self.filename), "cartesian_tx.nc")
        self.carta = self._loadcart(carta_file)
        self.carti = self._loadcart(carti_file)
        self.cartx = self._loadcart(cartx_file)

    @staticmethod
    def _interp_data(indata, full_grid, tie_grid, ds_name):
        """Interpolate data from tie point grid to full image grid."""
        from scipy.interpolate import RectBivariateSpline

        # Check if we are interpolating angles
        if "angle" in ds_name:
            # If we are interpolating the angles, we need to do so with the sine and cosine to prevent
            # interpolation artifacts such as values <0 or >360.
            indat = indata[:, ::-1]
            sin_angles = np.sin(np.radians(indat))
            cos_angles = np.cos(np.radians(indat))
            sin_interp = RectBivariateSpline(tie_grid[1], tie_grid[0], sin_angles)
            cos_interp = RectBivariateSpline(tie_grid[1], tie_grid[0], cos_angles)
            values_sin = sin_interp.ev(full_grid[1], full_grid[0])
            values_cos = cos_interp.ev(full_grid[1], full_grid[0])
            values = np.degrees(np.arctan2(values_sin, values_cos)) % 360
        else:
            # Otherwise, interpolate as normal.
            spl = RectBivariateSpline(tie_grid[1], tie_grid[0], indata[:, ::-1])
            values = spl.ev(full_grid[1], full_grid[0])
        return values


    def get_dataset(self, key, info):
        """Load a dataset."""
        if not key["view"].name.startswith(self.view[0]):
            return
        logger.debug("Reading %s.", key["name"])
        # Check if file_key is specified in the yaml
        file_key = info["file_key"].format(view=key["view"].name[0])

        variable = self.nc[file_key]
        l_step = self.nc.attrs.get("al_subsampling_factor", 1)
        c_step = self.nc.attrs.get("ac_subsampling_factor", 16)

        if key.get("resolution", 1000) == 500:
            l_step *= 2
            c_step *= 2

        if c_step != 1 or l_step != 1:
            logger.debug("Interpolating %s.", key["name"])
            # TODO: do it in cartesian coordinates ! pbs at date line and
            # possible
            tie_x = self.cartx["x_tx"].data[0, :][::-1]
            tie_y = self.cartx["y_tx"].data[:, 0]
            if key.get("resolution", 1000) == 500:
                full_x = self.carta["x_a" + self.view[0]].data
                full_y = self.carta["y_a" + self.view[0]].data
            else:
                full_x = self.carti["x_i" + self.view[0]].data
                full_y = self.carti["y_i" + self.view[0]].data

            variable = variable.fillna(0)
            variable.attrs["resolution"] = key.get("resolution", 1000)

            # Interpolate the data to the full image grid
            values = self._interp_data(variable.data,
                                       [full_x, full_y],
                                       [tie_x, tie_y],
                                       key["name"])

            variable = xr.DataArray(da.from_array(values, chunks=(CHUNK_SIZE, CHUNK_SIZE)),
                                    dims=["y", "x"], attrs=variable.attrs)

        variable.attrs["platform_name"] = self.platform_name
        variable.attrs["sensor"] = self.sensor

        if "units" not in variable.attrs:
            variable.attrs["units"] = "degrees"

        variable.attrs.update(key.to_dict())

        return variable

    @property
    def start_time(self):
        """Get the start time."""
        return dt.datetime.strptime(self.nc.attrs["start_time"], "%Y-%m-%dT%H:%M:%S.%fZ")

    @property
    def end_time(self):
        """Get the end time."""
        return dt.datetime.strptime(self.nc.attrs["stop_time"], "%Y-%m-%dT%H:%M:%S.%fZ")


class NCSLSTRFlag(BaseFileHandler):
    """File handler for flags."""

    def __init__(self, filename, filename_info, filetype_info):
        """Initialize the flag reader."""
        super(NCSLSTRFlag, self).__init__(filename, filename_info,
                                          filetype_info)
        self.nc = xr.open_dataset(self.filename,
                                  decode_cf=True,
                                  mask_and_scale=True,
                                  chunks={"columns": CHUNK_SIZE,
                                          "rows": CHUNK_SIZE})
        self.nc = self.nc.rename({"columns": "x", "rows": "y"})
        self.stripe = filename_info["stripe"]
        views = {"n": "nadir", "o": "oblique"}
        self.view = views[filename_info["view"]]
        # TODO: get metadata from the manifest file (xfdumanifest.xml)
        self.platform_name = PLATFORM_NAMES[filename_info["mission_id"]]
        self.sensor = "slstr"

    def get_dataset(self, key, info):
        """Load a dataset."""
        if (self.stripe != key["stripe"].name or
                self.view != key["view"].name):
            return
        logger.debug("Reading %s.", key["name"])
        file_key = info["file_key"].format(view=key["view"].name[0],
                                           stripe=key["stripe"].name)
        variable = self.nc[file_key]

        info = info.copy()
        info.update(variable.attrs)
        info.update(key.to_dict())
        info.update(dict(platform_name=self.platform_name,
                         sensor=self.sensor))

        variable.attrs = info
        return variable

    @property
    def start_time(self):
        """Get the start time."""
        return dt.datetime.strptime(self.nc.attrs["start_time"], "%Y-%m-%dT%H:%M:%S.%fZ")

    @property
    def end_time(self):
        """Get the end time."""
        return dt.datetime.strptime(self.nc.attrs["stop_time"], "%Y-%m-%dT%H:%M:%S.%fZ")



#!/usr/bin/env python
# Copyright (c) 2014-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Helper functions for satpy readers."""
from __future__ import annotations

import bz2
import logging
import os
import shutil
import tempfile
import warnings
from contextlib import closing, contextmanager
from io import BytesIO
from shutil import which
from subprocess import PIPE, Popen  # nosec

import numpy as np
import pyproj
import xarray as xr
from pyresample.geometry import AreaDefinition
from upath import UPath

from satpy import config
from satpy.readers.core.remote import FSFile
from satpy.utils import get_legacy_chunk_size

LOGGER = logging.getLogger(__name__)
CHUNK_SIZE = get_legacy_chunk_size()


def np2str(value: np.ndarray) -> str:
    """Convert an `numpy.string_` to str.

    Args:
        value: scalar or 1-element numpy array to convert

    Raises:
        ValueError: if value is array larger than 1-element, or it is not of
                    type `numpy.string_` or it is not a numpy array

    """
    if hasattr(value, "dtype") and \
            issubclass(value.dtype.type, (np.str_, np.bytes_, np.object_)) \
            and value.size == 1:
        value = value.item()
        if not isinstance(value, str):
            # python 3 - was scalar numpy array of bytes
            # otherwise python 2 - scalar numpy array of 'str'
            value = value.decode()
        return value
    else:
        raise ValueError("Array is not a string type or is larger than 1")


def _get_geostationary_height(geos_area):
    params = geos_area.crs.coordinate_operation.params
    h_param = [p for p in params if "satellite height" in p.name.lower()][0]
    return h_param.value


def _get_geostationary_reference_longitude(geos_area):
    params = geos_area.crs.coordinate_operation.params
    lon_0_params = [p for p in params if "longitude of natural origin" in p.name.lower()]
    if not lon_0_params:
        return 0
    elif len(lon_0_params) != 1:
        raise ValueError("Not sure how to get reference longitude "
                         "information from AreaDefinition.")
    return lon_0_params[0].value


def _get_geostationary_semi_axes(geos_area):
    from pyresample.utils import proj4_radius_parameters
    return proj4_radius_parameters(geos_area.crs)


def get_geostationary_angle_extent(geos_area):
    """Get the max earth (vs space) viewing angles in x and y."""
    # TODO: take into account sweep_axis_angle parameter
    a, b = _get_geostationary_semi_axes(geos_area)
    h = _get_geostationary_height(geos_area)
    req = float(a) / 1000
    rp = float(b) / 1000
    h = float(h) / 1000 + req

    # compute some constants
    aeq = 1 - req ** 2 / (h ** 2)
    ap_ = 1 - rp ** 2 / (h ** 2)

    # generate points around the north hemisphere in satellite projection
    # make it a bit smaller so that we stay inside the valid area
    xmax = np.arccos(np.sqrt(aeq))
    ymax = np.arccos(np.sqrt(ap_))
    return xmax, ymax


def get_geostationary_mask(area, chunks=None):
    """Compute a mask of the earth's shape as seen by a geostationary satellite.

    Args:
        area (pyresample.geometry.AreaDefinition) : Corresponding area
                                                    definition
        chunks (int or tuple): Chunk size for the 2D array that is generated.

    Returns:
        Boolean mask, True inside the earth's shape, False outside.

    """
    # Compute projection coordinates at the earth's limb
    h = _get_geostationary_height(area)
    xmax, ymax = get_geostationary_angle_extent(area)
    xmax *= h
    ymax *= h

    # Compute projection coordinates at the centre of each pixel
    x, y = area.get_proj_coords(chunks=chunks or CHUNK_SIZE)

    # Compute mask of the earth's elliptical shape
    return ((x / xmax) ** 2 + (y / ymax) ** 2) <= 1


def _lonlat_from_geos_angle(x, y, geos_area):
    """Get lons and lats from x, y in projection coordinates."""
    a, b = _get_geostationary_semi_axes(geos_area)
    h = _get_geostationary_height(geos_area)
    lon_0 = _get_geostationary_reference_longitude(geos_area)
    h__ = float(h + a) / 1000
    b__ = (a / float(b)) ** 2

    sd = np.sqrt((h__ * np.cos(x) * np.cos(y)) ** 2 -
                 (np.cos(y) ** 2 + b__ * np.sin(y) ** 2) *
                 (h__ ** 2 - (float(a) / 1000) ** 2))
    # sd = 0

    sn = (h__ * np.cos(x) * np.cos(y) - sd) / (np.cos(y) ** 2 + b__ * np.sin(y) ** 2)
    s1 = h__ - sn * np.cos(x) * np.cos(y)
    s2 = sn * np.sin(x) * np.cos(y)
    s3 = -sn * np.sin(y)
    sxy = np.sqrt(s1 ** 2 + s2 ** 2)

    lons = np.rad2deg(np.arctan2(s2, s1)) + lon_0
    lats = np.rad2deg(-np.arctan2(b__ * s3, sxy))

    return lons, lats


def get_geostationary_bounding_box(geos_area, nb_points=50):
    """Get the bbox in lon/lats of the valid pixels inside *geos_area*.

    Args:
      geos_area: The geostationary area to analyse.
      nb_points: Number of points on the polygon

    """
    xmax, ymax = get_geostationary_angle_extent(geos_area)
    h = _get_geostationary_height(geos_area)

    # generate points around the north hemisphere in satellite projection
    # make it a bit smaller so that we stay inside the valid area
    x = np.cos(np.linspace(-np.pi, 0, nb_points // 2)) * (xmax - 0.001)
    y = -np.sin(np.linspace(-np.pi, 0, nb_points // 2)) * (ymax - 0.001)

    # clip the projection coordinates to fit the area extent of geos_area
    ll_x, ll_y, ur_x, ur_y = (np.array(geos_area.area_extent) /
                              float(h))

    x = np.clip(np.concatenate([x, x[::-1]]), min(ll_x, ur_x), max(ll_x, ur_x))
    y = np.clip(np.concatenate([y, -y]), min(ll_y, ur_y), max(ll_y, ur_y))

    return _lonlat_from_geos_angle(x, y, geos_area)


def get_sub_area(area, xslice, yslice):
    """Apply slices to the area_extent and size of the area."""
    new_area_extent = ((area.pixel_upper_left[0] +
                        (xslice.start - 0.5) * area.pixel_size_x),
                       (area.pixel_upper_left[1] -
                        (yslice.stop - 0.5) * area.pixel_size_y),
                       (area.pixel_upper_left[0] +
                        (xslice.stop - 0.5) * area.pixel_size_x),
                       (area.pixel_upper_left[1] -
                        (yslice.start - 0.5) * area.pixel_size_y))

    return AreaDefinition(area.area_id, area.name,
                          area.proj_id, area.crs,
                          xslice.stop - xslice.start,
                          yslice.stop - yslice.start,
                          new_area_extent)


def unzip_file(filename: str | FSFile, prefix=None):
    """Unzip the local/remote file ending with 'bz2'.

    Args:
        filename: The local/remote file to unzip.
        prefix (str, Optional): If file is one of many segments of data, prefix random filename
                                for correct sorting. This is normally the segment number.

    Returns:
        (pathlib.Path, None) Temporary filename path for decompressed file if available.

    """
    if isinstance(filename, str):
        return _unzip_local_file(filename, prefix=prefix)
    elif isinstance(filename, FSFile):
        return _unzip_FSFile(filename, prefix=prefix)


def _unzip_local_file(filename: str, prefix=None):
    """Unzip the file ending with 'bz2'. Initially with pbzip2 if installed or bz2.

    Args:
        filename: The file to unzip.
        prefix (str, Optional): If file is one of many segments of data, prefix random filename
                                for correct sorting. This is normally the segment number.

    Returns:
        (pathlib.Path, None) Temporary filename path for decompressed file if available.

    """
    if not os.fspath(filename).endswith("bz2"):
        return None
    fdn, tmpfilepath = tempfile.mkstemp(prefix=prefix,
                                        dir=config["tmp_dir"])
    LOGGER.info("Using temp file for BZ2 decompression: %s", tmpfilepath)
    # check pbzip2 status
    pbzip2 = _unzip_with_pbzip(filename, tmpfilepath, fdn)
    if pbzip2 is not None:
        return pbzip2
    # Otherwise, fall back to the original method bz2
    content = _unzip_with_bz2(filename, tmpfilepath)
    return _write_uncompressed_file(content, fdn, filename, tmpfilepath)


def _unzip_with_pbzip(filename, tmpfilepath, fdn):
    # try pbzip2
    pbzip = which("pbzip2")
    if pbzip is None:
        return None
    # Run external pbzip2
    n_thr = os.environ.get("OMP_NUM_THREADS")
    if n_thr:
        runner = [pbzip,
                  "-dc",
                  "-p" + str(n_thr),
                  filename]
    else:
        runner = [pbzip,
                  "-dc",
                  filename]
    p = Popen(runner, stdout=PIPE, stderr=PIPE)  # nosec
    stdout = BytesIO(p.communicate()[0])
    status = p.returncode
    if status != 0:
        raise IOError("pbzip2 error '%s', failed, status=%d"
                      % (filename, status))
    with closing(os.fdopen(fdn, "wb")) as ofpt:
        try:
            stdout.seek(0)
            shutil.copyfileobj(stdout, ofpt)
        except IOError:
            LOGGER.debug("Failed to read bzipped file %s", str(filename))
            os.remove(tmpfilepath)
            raise
    return tmpfilepath


def _unzip_with_bz2(filename, tmpfilepath):
    with bz2.BZ2File(filename) as bz2file:
        try:
            content = bz2file.read()
        except IOError:
            LOGGER.debug("Failed to unzip bzipped file %s", str(filename))
            os.remove(tmpfilepath)
            raise
    return content


def _write_uncompressed_file(content, fdn, filename, tmpfilepath):
    with closing(os.fdopen(fdn, "wb")) as ofpt:
        try:
            ofpt.write(content)
        except IOError:
            LOGGER.debug("Failed to write uncompressed file %s", str(filename))
            os.remove(tmpfilepath)
            return None
    return tmpfilepath


def _unzip_FSFile(filename: FSFile, prefix=None):
    """Open and Unzip remote FSFile ending with 'bz2'.

    Args:
        filename: The FSFile to unzip.
        prefix (str, Optional): If file is one of many segments of data, prefix random filename
                                for correct sorting. This is normally the segment number.

    Returns:
        (os.Pathlike, None) Temporary filename path for decompressed file if available.

    """
    fdn, tmpfilepath = tempfile.mkstemp(prefix=prefix,
                                        dir=config["tmp_dir"])
    # open file
    content = filename.open().read()
    # unzip file if zipped (header start with hex 425A68)
    if content.startswith(bytes.fromhex("425A68")):
        content = bz2.decompress(content)

    return _write_uncompressed_file(content, fdn, filename, tmpfilepath)


@contextmanager
def unzip_context(filename):
    """Context manager for decompressing a .bz2 file on the fly.

    Uses `unzip_file`. Removes the uncompressed file on exit of the context manager.

    Returns: the filename of the uncompressed file or of the original file if it was not
    compressed.

    """
    unzipped = unzip_file(filename)
    if unzipped is not None:
        yield unzipped
        os.remove(unzipped)
    else:
        yield filename


@contextmanager
def generic_open(filename, *args, **kwargs):
    """Context manager for opening either a regular file or a compressed file.

    Yields an open file-like object.
    """
    path = _to_upath(filename)

    new_kwargs = kwargs.copy()
    new_kwargs["compression"] = "infer"
    with path.open(*args, **new_kwargs) as fp:
        yield fp


def _to_upath(filename: FSFile | str | os.PathLike | UPath) -> UPath:
    """Convert to UPath."""
    if isinstance(filename, UPath):
        return filename

    if hasattr(filename, "to_upath"):
        # satpy.readers.core.remote.FSFile instance
        return filename.to_upath()

    if isinstance(filename, (str, os.PathLike)):
        return UPath(filename, protocol="file")

    warnings.warn(f"Converting {type(filename)} to UPath.", UserWarning)
    return UPath(filename)



def fromfile(filename, dtype, count=1, offset=0):
    """Read the numpy array from a (remote or local) file using a buffer.

    Note:
        This function relies on the :func:`generic_open` context manager to read a file remotely.

    Args:
        filename: Either the name of the file to read or a :class:`satpy.readers.core.remote` object.
        dtype: The data type of the numpy array
        count (Optional): Number of items to read. Default ``1``
        offset (Optional): Starting point for reading the buffer from.
                           Default ``0``

    Returns:
        The content of the filename as a numpy array with the given data type.
    """
    with generic_open(filename, mode="rb") as istream:
        istream.seek(offset)
        content = np.frombuffer(istream.read(dtype.itemsize * count), dtype=dtype, count=count)
    return content


def bbox(img):
    """Find the bounding box around nonzero elements in the given array.

    Copied from https://stackoverflow.com/a/31402351/5703449 .

    Returns:
        rowmin, rowmax, colmin, colmax

    """
    rows = np.any(img, axis=1)
    cols = np.any(img, axis=0)
    rmin, rmax = np.where(rows)[0][[0, -1]]
    cmin, cmax = np.where(cols)[0][[0, -1]]

    return rmin, rmax, cmin, cmax


def get_earth_radius(lon, lat, a, b):
    """Compute radius of the earth ellipsoid at the given longitude and latitude.

    Args:
        lon: Geodetic longitude (degrees)
        lat: Geodetic latitude (degrees)
        a: Semi-major axis of the ellipsoid (meters)
        b: Semi-minor axis of the ellipsoid (meters)

    Returns:
        Earth Radius (meters)

    """
    geocent = pyproj.CRS.from_dict({"proj": "geocent", "a": a, "b": b, "units": "m"})
    latlong = pyproj.CRS.from_dict({"proj": "latlong", "a": a, "b": b, "units": "m"})
    transformer = pyproj.Transformer.from_crs(latlong, geocent)
    x, y, z = transformer.transform(lon, lat, 0.0)
    return np.sqrt(x ** 2 + y ** 2 + z ** 2)


def reduce_mda(mda, max_size=100):
    """Recursively remove arrays with more than `max_size` elements from the given metadata dictionary."""
    reduced = {}
    for key, val in mda.items():
        if isinstance(val, dict):
            reduced[key] = reduce_mda(val, max_size)
        elif not (isinstance(val, np.ndarray) and val.size > max_size):
            reduced[key] = val
    return reduced


def get_user_calibration_factors(band_name, correction_dict):
    """Retrieve radiance correction factors from user-supplied dict."""
    if band_name in correction_dict:
        try:
            slope = correction_dict[band_name]["slope"]
            offset = correction_dict[band_name]["offset"]
        except KeyError:
            raise KeyError("Incorrect correction factor dictionary. You must "
                           "supply 'slope' and 'offset' keys.")
    else:
        # If coefficients not present, warn user and use slope=1, offset=0
        warnings.warn(
            "WARNING: You have selected radiance correction but "
            " have not supplied coefficients for channel " + band_name,
            stacklevel=2
        )
        return 1., 0.

    return slope, offset


def apply_rad_correction(data, slope, offset):
    """Apply GSICS-like correction factors to radiance data."""
    data = (data - offset) / slope
    return data


def get_array_date(scn_data, utc_date=None):
    """Get start time from a channel data array."""
    if utc_date is None:
        try:
            utc_date = scn_data.attrs["start_time"]
        except KeyError:
            try:
                utc_date = scn_data.attrs["scheduled_time"]
            except KeyError:
                raise KeyError("Scene has no start_time "
                               "or scheduled_time attribute.")
    return utc_date


def apply_earthsun_distance_correction(reflectance, utc_date=None):
    """Correct reflectance data to account for changing Earth-Sun distance."""
    from pyorbital.astronomy import sun_earth_distance_correction
    utc_date = get_array_date(reflectance, utc_date)
    sun_earth_dist = sun_earth_distance_correction(utc_date)

    reflectance.attrs["sun_earth_distance_correction_applied"] = True
    reflectance.attrs["sun_earth_distance_correction_factor"] = sun_earth_dist
    with xr.set_options(keep_attrs=True):
        reflectance = reflectance * reflectance.dtype.type(sun_earth_dist * sun_earth_dist)
    return reflectance


def remove_earthsun_distance_correction(reflectance, utc_date=None):
    """Remove the sun-earth distance correction."""
    from pyorbital.astronomy import sun_earth_distance_correction
    utc_date = get_array_date(reflectance, utc_date)
    sun_earth_dist = sun_earth_distance_correction(utc_date)

    reflectance.attrs["sun_earth_distance_correction_applied"] = False
    reflectance.attrs["sun_earth_distance_correction_factor"] = sun_earth_dist
    with xr.set_options(keep_attrs=True):
        reflectance = reflectance / reflectance.dtype.type(sun_earth_dist * sun_earth_dist)
    return reflectance


class _CalibrationCoefficientParser:
    """Parse user-defined calibration coefficients."""

    def __init__(self, coefs, default="nominal"):
        """Initialize the parser."""
        if default not in coefs:
            raise KeyError("Need at least default coefficients")
        self.coefs = coefs
        self.default = default

    def parse(self, calib_wishlist):
        """Parse user's calibration wishlist."""
        if calib_wishlist is None:
            return self._get_coefs_set(self.default)
        elif isinstance(calib_wishlist, str):
            return self._get_coefs_set(calib_wishlist)
        elif isinstance(calib_wishlist, dict):
            return self._parse_dict(calib_wishlist)
        raise TypeError(
            f"Unsupported wishlist type. Expected dict/str, "
            f"got {type(calib_wishlist)}"
        )

    def _parse_dict(self, calib_wishlist):
        calib_wishlist = self._flatten_multi_channel_keys(calib_wishlist)
        return self._replace_calib_mode_with_actual_coefs(calib_wishlist)

    def _flatten_multi_channel_keys(self, calib_wishlist):
        flat = {}
        for channels, coefs in calib_wishlist.items():
            if self._is_multi_channel(channels):
                flat.update({channel: coefs for channel in channels})
            else:
                flat[channels] = coefs
        return flat

    def _is_multi_channel(self, key):
        return isinstance(key, tuple)

    def _replace_calib_mode_with_actual_coefs(self, calib_wishlist):
        res = {}
        for channel in self.coefs[self.default]:
            mode_or_coefs = calib_wishlist.get(channel, self.default)
            coefs = self._get_coefs(mode_or_coefs, channel)
            if coefs:
                res[channel] = coefs
        return res

    def _get_coefs(self, mode_or_coefs, channel):
        if self._is_mode(mode_or_coefs):
            return self._get_coefs_by_mode(mode_or_coefs, channel)
        return _make_coefs(mode_or_coefs, "external")

    def _is_mode(self, mode_or_coefs):
        return isinstance(mode_or_coefs, str)

    def _get_coefs_by_mode(self, mode, channel):
        coefs_set = self._get_coefs_set(mode)
        return coefs_set.get(channel, None)

    def _get_coefs_set(self, mode):
        try:
            return {
                channel: _make_coefs(coefs, mode)
                for channel, coefs in self.coefs[mode].items()
            }
        except KeyError:
            modes = list(self.coefs.keys())
            raise KeyError(f"Unknown calibration mode: {mode}. Choose one of {modes}")

    def get_calib_mode(self, calib_wishlist, channel):
        """Get desired calibration mode for the given channel."""
        if isinstance(calib_wishlist, str):
            return calib_wishlist
        elif isinstance(calib_wishlist, dict):
            flat = self._flatten_multi_channel_keys(calib_wishlist)
            return flat[channel]


class CalibrationCoefficientPicker:
    """Helper for choosing coefficients out of multiple options.

    Example: Three sets of coefficients are available (nominal, meirink, gsics).
    A user wants to calibrate

        - channel 1 with "meirink"
        - channels 2/3 with "gsics"
        - channel 4 with custom coefficients
        - remaining channels with nominal coefficients

    1. Users provide a wishlist via ``reader_kwargs``

    .. code-block:: python

        calib_wishlist = {
            "ch1": "meirink",
            ("ch2", "ch3"): "gsics",
            "ch4": {"mygain": 123},
        }
        # Also possible: Same mode for all channels via
        # calib_wishlist = "gsics"

    2. Readers provide a dictionary with all available coefficients

    .. code-block:: python

        coefs = {
            "nominal": {
                "ch1": 1.0,
                "ch2": 2.0,
                "ch3": 3.0,
                "ch4": 4.0,
                "ch5": 5.0,
            },
            "meirink": {
                "ch1": 1.1,
            },
            "gsics": {
                "ch2": 2.2,
                # ch3 coefficients are missing
            }
        }

    3. Raders make queries to get the desired coefficients:

    .. code-block:: python

        >>> from satpy.readers.core.utils import CalibrationCoefficientPicker
        >>> picker = CalibrationCoefficientPicker(coefs, calib_wishlist)
        >>> picker.get_coefs("ch1")
        {"coefs": 1.0, "mode": "meirink"}
        >>> picker.get_coefs("ch2")
        {"coefs": 2.2, "mode": "gsics"}
        >>> picker.get_coefs("ch3")
        KeyError: 'No gsics calibration coefficients for ch3'
        >>> picker.get_coefs("ch4")
        {"coefs": {"mygain": 123}, "mode": "external"}
        >>> picker.get_coefs("ch5")
        {"coefs": 5.0, "mode": "nominal"}

    4. Fallback to nominal coefficients for ch3:

    .. code-block:: python

        >>> picker = CalibrationCoefficientPicker(coefs, calib_wishlist, fallback="nominal")
        >>> picker.get_coefs("ch3")
        WARNING No gsics calibration coefficients for ch3. Falling back to nominal.
        {"coefs": 3.0, "mode": "nominal"}

    """

    def __init__(self, coefs, calib_wishlist, default="nominal", fallback=None):
        """Initialize the coefficient picker.

        Args:
            coefs (dict): One set of calibration coefficients for each
                calibration mode. The actual coefficients can be of any type
                (reader-specific).
            calib_wishlist (str or dict): Desired calibration coefficients. Use a
                dictionary to specify channel-specific coefficients. Use a
                string to specify one mode for all channels.
            default (str): Default coefficients to be used if nothing was
                specified in the calib_wishlist. Default: "nominal".
            fallback (str): Fallback coefficients if the desired coefficients
                are not available for some channel. By default, an exception is
                raised if coefficients are missing.
        """
        if fallback and fallback not in coefs:
            raise KeyError("No fallback calibration coefficients")
        self.coefs = coefs
        self.calib_wishlist = calib_wishlist
        self.default = default
        self.fallback = fallback
        self.parser = _CalibrationCoefficientParser(coefs, default)
        self.parsed_wishlist = self.parser.parse(calib_wishlist)

    def get_coefs(self, channel):
        """Get calibration coefficients for the given channel.

        Args:
            channel (str): Channel name

        Returns:
            dict: Calibration coefficients and mode (for transparency, in case
                  the picked coefficients differ from the wishlist).
        """
        try:
            return self.parsed_wishlist[channel]
        except KeyError:
            mode = self.parser.get_calib_mode(self.calib_wishlist, channel)
            if self.fallback:
                LOGGER.warning(
                    f"No {mode} calibration coefficients for {channel}. "
                    f"Falling back to {self.fallback}."
                )
                return _make_coefs(self.coefs[self.fallback][channel],
                                   self.fallback)
            raise KeyError(f"No {mode} calibration coefficients for {channel}")


def _make_coefs(coefs, mode):
    return {"coefs": coefs, "mode": mode}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2016-2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Reader for AMSR2 L1B files in HDF5 format."""

from satpy.readers.core.hdf5 import HDF5FileHandler


class AMSR2L1BFileHandler(HDF5FileHandler):
    """File handler for AMSR2 l1b."""

    def get_metadata(self, ds_id, ds_info):
        """Get the metadata."""
        var_path = ds_info["file_key"]
        info = getattr(self[var_path], "attrs", {})
        info.update(ds_info)
        info.update({
            "shape": self.get_shape(ds_id, ds_info),
            "units": self[var_path + "/attr/UNIT"],
            "platform_name": self["/attr/PlatformShortName"],
            "sensor": self["/attr/SensorShortName"].lower(),
            "start_orbit": int(self["/attr/StartOrbitNumber"]),
            "end_orbit": int(self["/attr/StopOrbitNumber"]),
        })
        info.update(ds_id.to_dict())
        return info

    def get_shape(self, ds_id, ds_info):
        """Get output shape of specified dataset."""
        var_path = ds_info["file_key"]
        shape = self[var_path + "/shape"]
        if ((ds_info.get("standard_name") == "longitude" or ds_info.get("standard_name") == "latitude") and
                ds_id["resolution"] == 10000):
            return shape[0], int(shape[1] / 2)
        return shape

    def get_dataset(self, ds_id, ds_info):
        """Get output data and metadata of specified dataset."""
        var_path = ds_info["file_key"]
        fill_value = ds_info.get("fill_value", 65535)
        metadata = self.get_metadata(ds_id, ds_info)

        data = self[var_path]
        if ((ds_info.get("standard_name") == "longitude" or
             ds_info.get("standard_name") == "latitude") and
                ds_id["resolution"] == 10000):
            # FIXME: Lower frequency channels need CoRegistration parameters applied
            data = data[:, ::2] * self[var_path + "/attr/SCALE FACTOR"]
        else:
            data = data * self[var_path + "/attr/SCALE FACTOR"]
        data = data.where(data != fill_value)
        data.attrs.update(metadata)
        return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2020 Satpy developers
#
# satpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# satpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with satpy.  If not, see <http://www.gnu.org/licenses/>.

"""The vii_utils reader tests package."""

import unittest

import satpy.readers.core.vii

# Constants to be tested
C1 = 1.191062e+8
C2 = 1.4387863e+4
TIE_POINTS_FACTOR = 8
SCAN_ALT_TIE_POINTS = 4
MEAN_EARTH_RADIUS = 6371008.7714


class TestViiUtils(unittest.TestCase):
    """Test the vii_utils module."""

    def test_constants(self):
        """Test the constant values."""
        # Test the value of the constants
        assert satpy.readers.core.vii.C1 == C1
        assert satpy.readers.core.vii.C2 == C2
        assert satpy.readers.core.vii.TIE_POINTS_FACTOR == TIE_POINTS_FACTOR
        assert satpy.readers.core.vii.SCAN_ALT_TIE_POINTS == SCAN_ALT_TIE_POINTS
        assert satpy.readers.core.vii.MEAN_EARTH_RADIUS == MEAN_EARTH_RADIUS
